from django.db.models import Exists, OuterRef, Q
from wbcore import filters as wb_filters
from wbcore.contrib.directory.models import Entry, Person
from wbcore.contrib.pandas.filterset import PandasFilterSetMixin
from wbcore.filters.defaults import year_to_date_range
from wbcore.utils.date import current_financial_quarter
from wbcrm.models.accounts import Account, AccountRole
from wbfdm.models import Classification, ClassificationGroup
from wbfdm.preferences import get_default_classification_group

from wbportfolio.filters.transactions.mixins import OppositeSharesFieldMethodMixin
from wbportfolio.models import Custodian, Product, ProductGroup, Register
from wbportfolio.models.transactions.claim import Claim, ClaimGroupbyChoice
from wbportfolio.preferences import get_monthly_nnm_target


class CustomerAPIFilter(wb_filters.FilterSet):
    external_id = wb_filters.CharFilter(label="External ID", lookup_expr="exact")

    class Meta:
        model = Claim
        fields = []


class CustomerClaimFilter(wb_filters.FilterSet):
    class Meta:
        model = Claim
        fields = {"date": ["gte", "lte", "exact"], "status": ["exact"], "product": ["exact"], "bank": ["exact"]}


def get_entry_queryset(request):
    """
    Show only entry which have a account role the user can see
    """
    queryset = Entry.objects.all()
    if request:
        roles = AccountRole.objects.filter_for_user(request.user)
        queryset = queryset.filter(account_roles__in=roles)
    return queryset


class CommissionBaseFilterSet(wb_filters.FilterSet):
    account_owner = wb_filters.ModelMultipleChoiceFilter(
        label="Account Owner",
        queryset=Entry.objects.all(),
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
        filter_params={"with_account": True},
        method="filter_account_owner",
    )

    not_account_owner = wb_filters.ModelMultipleChoiceFilter(
        column_field_name="account_owner",
        lookup_icon="!=",
        lookup_label="Not Equals",
        queryset=Entry.objects.all(),
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
        filter_params={"with_account": True},
        method="filter_not_owner",
    )

    product = wb_filters.ModelMultipleChoiceFilter(
        label="Product",
        queryset=Product.objects.all(),
        endpoint=Product.get_representation_endpoint(),
        value_key=Product.get_representation_value_key(),
        label_key=Product.get_representation_label_key(),
    )
    product__parent = wb_filters.ModelMultipleChoiceFilter(
        label="Product Group",
        queryset=ProductGroup.objects.all(),
        endpoint=ProductGroup.get_representation_endpoint(),
        value_key=ProductGroup.get_representation_value_key(),
        label_key=ProductGroup.get_representation_label_key(),
    )

    account = wb_filters.ModelChoiceFilter(
        label="Account",
        method="filter_account",
        queryset=Account.objects.all(),
        endpoint=Account.get_representation_endpoint(),
        value_key=Account.get_representation_value_key(),
        label_key=Account.get_representation_label_key(),
    )

    manager_role = wb_filters.ModelChoiceFilter(
        label="Manager Role",
        queryset=get_entry_queryset,
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
        filter_params={"with_account_role": True},
        method="filter_manager_role",
    )

    product__classifications = wb_filters.ModelMultipleChoiceFilter(
        label="Classifications",
        queryset=Classification.objects.all(),
        endpoint=Classification.get_representation_endpoint(),
        value_key=Classification.get_representation_value_key(),
        label_key=Classification.get_representation_label_key(),
        filter_params={"instrument_type_key": "product"},
    )

    def filter_account(self, queryset, name, value):
        if value:
            return queryset.filter(account__in=value.get_descendants(include_self=True).values("id"))
        return queryset

    def filter_account_owner(self, queryset, name, value):
        if value:
            return queryset.filter(account__in=Account.get_accounts_for_customer(value))
        return queryset

    def filter_not_owner(self, queryset, name, value):
        if value:
            return queryset.exclude(account__in=Account.get_accounts_for_customer(value))
        return queryset

    def filter_manager_role(self, queryset, name, value):
        if value:
            return queryset.filter(account__in=Account.get_managed_accounts_for_entry(value).values("id"))
        return queryset


class ClaimFilter(OppositeSharesFieldMethodMixin, CommissionBaseFilterSet):
    # we have to redefine the mixin fields because django_filters does not allow class extension with mixin
    opposite_shares = wb_filters.NumberFilter(
        field_name="shares",
        label="Opposite Shares",
        method="filter_opposite_shares",
        lookup_icon="±",
        lookup_label="Opposite",
    )
    opposite_approximate_shares = wb_filters.NumberFilter(
        field_name="shares",
        label="Opposite Approximite Shares",
        method="filter_opposite_approximate_shares",
        lookup_icon="≈±",
        lookup_label="Opposite Approximate (+- 10%)",
    )

    pending_approval = wb_filters.BooleanFilter(label="Pending Approval", method="boolean_pending_approval")
    linked_trade = wb_filters.BooleanFilter(method="filter_trade_isnotnull", label="Trade Linked")
    in_charge_of_customer = wb_filters.ModelChoiceFilter(
        label="In Charge of Customer",
        method="filter_in_charge_of_customer",
        queryset=Person.objects.all(),
        endpoint=Person.get_representation_endpoint(),
        value_key=Person.get_representation_value_key(),
        label_key=Person.get_representation_label_key(),
    )

    # Trade related filter fields
    trade__custodian = wb_filters.ModelChoiceFilter(
        label="Trade Custodian",
        queryset=Custodian.objects.all(),
        endpoint=Custodian.get_representation_endpoint(),
        value_key=Custodian.get_representation_value_key(),
        label_key=Custodian.get_representation_label_key(),
    )
    trade__register = wb_filters.ModelChoiceFilter(
        label="Trade Register",
        queryset=Register.objects.all(),
        endpoint=Register.get_representation_endpoint(),
        value_key=Register.get_representation_value_key(),
        label_key=Register.get_representation_label_key(),
    )

    trade_comment = wb_filters.CharFilter(lookup_expr="icontains", label="Trade Comment")

    last_nav = wb_filters.NumberFilter(field_name="last_nav", lookup_expr="exact", label="Last Nav")
    last_nav__lte = wb_filters.NumberFilter(field_name="last_nav", lookup_expr="lte", label="Last Nav")
    last_nav__gte = wb_filters.NumberFilter(field_name="last_nav", lookup_expr="gte", label="Last Nav")

    total_value = wb_filters.NumberFilter(field_name="total_value", lookup_expr="exact", label="Total Value")
    total_value__lte = wb_filters.NumberFilter(field_name="total_value", lookup_expr="lte", label="Total Value")
    total_value__gte = wb_filters.NumberFilter(field_name="total_value", lookup_expr="gte", label="Total Value")

    total_value_usd = wb_filters.NumberFilter(
        field_name="total_value_usd", lookup_expr="exact", label="Total Value (USD)"
    )
    total_value_usd__lte = wb_filters.NumberFilter(
        field_name="total_value_usd", lookup_expr="lte", label="Total Value (USD)"
    )
    total_value_usd__gte = wb_filters.NumberFilter(
        field_name="total_value_usd", lookup_expr="gte", label="Total Value (USD)"
    )

    def filter_in_charge_of_customer(self, queryset, name, value):
        if value:
            return queryset.filter(
                Q(claimant__relationship_managers=value)
                | Q(root_account__customers__entry__relationship_managers=value)  # TODO optimize this
            ).distinct("id")
        return queryset

    def filter_trade_isnotnull(self, queryset, name, value):
        if value is True:
            return queryset.filter(trade__isnull=False)
        elif value is False:
            return queryset.filter(trade__isnull=True)
        return queryset

    def boolean_pending_approval(self, queryset, name, value):
        if value is True:
            return queryset.filter(status=Claim.Status.PENDING, account__isnull=False, trade__isnull=False)
        elif value is False:
            return queryset.filter(status__in=[Claim.Status.APPROVED])
        return queryset

    class Meta:
        model = Claim
        fields = {
            "date": ["gte", "lte", "exact"],
            "status": ["exact"],
            # "trade": ["exact"],
            "bank": ["exact", "icontains"],
            "claimant": ["exact"],
            "shares": ["gte", "lte", "exact"],
            "reference": ["exact", "icontains"],
            "creator": ["exact"],
        }


class ClaimGroupByFilter(CommissionBaseFilterSet):
    pending_approval = in_charge_of_customer = linked_trade = None

    only_new_customer = wb_filters.BooleanFilter(method="filter_only_new_customer", label="Only new customers")

    group_by = wb_filters.ChoiceFilter(
        label="Group By",
        choices=ClaimGroupbyChoice.choices(),
        initial=ClaimGroupbyChoice.PRODUCT.name,
        method=lambda queryset, label, value: queryset,
        required=True,
        clearable=False,
    )

    date = wb_filters.DateRangeFilter(
        label="Date Range",
        method=lambda queryset, label, value: queryset,
        required=True,
        clearable=False,
        initial=year_to_date_range,
    )

    groupby_classification_group = wb_filters.ModelChoiceFilter(
        initial=lambda k, v, f: get_default_classification_group().id,
        method=lambda queryset, label, value: queryset,
        label="Group by Classification Group",
        queryset=ClassificationGroup.objects.all(),
        endpoint=ClassificationGroup.get_representation_endpoint(),
        value_key=ClassificationGroup.get_representation_value_key(),
        label_key=ClassificationGroup.get_representation_label_key(),
        depends_on=[{"field": "group_by", "options": {"activates_on": [ClaimGroupbyChoice.CLASSIFICATION.name]}}],
    )

    def filter_only_new_customer(self, queryset, name, value):
        if value:
            past_accounts = Account.all_objects.annotate(
                has_past_claims=Exists(
                    queryset.filter(date__lt=self.view.start_date, account__tree_id=OuterRef("tree_id"))
                )
            ).filter(has_past_claims=True)
            return queryset.exclude(account__in=past_accounts)
        return queryset

    class Meta:
        model = Claim
        fields = {}


class ConsolidatedTradeSummaryTableFilterSet(PandasFilterSetMixin, ClaimGroupByFilter):
    class Meta:
        model = Claim
        fields = {}
        df_fields = {
            "sum_shares_start__lte": wb_filters.NumberFilter(
                label="Sum Shares Start", lookup_expr="gte", field_name="sum_shares_start"
            ),
            "sum_shares_start__gte": wb_filters.NumberFilter(
                label="Sum Shares Start", lookup_expr="lte", field_name="sum_shares_start"
            ),
            "sum_shares_end__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Sum Shares End", field_name="sum_shares_end"
            ),
            "sum_shares_diff__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Difference Shares", field_name="sum_shares_diff"
            ),
            "sum_aum_start__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Sum AUM Start", field_name="sum_aum_start"
            ),
            "sum_aum_end__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Sum AUM End", field_name="sum_aum_end"
            ),
            "sum_aum_diff__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Nominal difference", field_name="sum_aum_diff"
            ),
            "sum_aum_perf__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Pourcent difference", field_name="sum_aum_perf"
            ),
            "sum_shares_end__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Sum Shares End", field_name="sum_shares_end"
            ),
            "sum_shares_diff__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Difference Shares", field_name="sum_shares_diff"
            ),
            "sum_aum_start__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Sum AUM Start", field_name="sum_aum_start"
            ),
            "sum_aum_end__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Sum AUM End", field_name="sum_aum_end"
            ),
            "sum_aum_diff__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Nominal difference", field_name="sum_aum_diff"
            ),
            "sum_aum_perf__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Pourcent difference", field_name="sum_aum_perf"
            ),
        }


class DistributionNNMChartFilter(ClaimGroupByFilter):
    percent = wb_filters.BooleanFilter(
        method="fake_filter",
        initial=False,
        required=True,
        help_text="True if the value are displayed in percentage of the initial total AUM",
        label="Show percentage",
    )

    class Meta:
        model = Claim
        fields = {}


class CumulativeNNMChartFilter(ClaimGroupByFilter):
    groupby_classification_group = group_by = None

    hide_projected_monthly_nnm_target = wb_filters.BooleanFilter(
        initial=True, required=True, label="Hide Target line", method=lambda queryset, label, value: queryset
    )
    projected_monthly_nnm_target = wb_filters.NumberFilter(
        initial=lambda k, v, f: get_monthly_nnm_target(),
        method=lambda queryset, label, value: queryset,
        label="Projected Monthly NNM Target",
    )

    class Meta:
        model = Claim
        fields = {}


class ProfitAndLossPandasFilter(PandasFilterSetMixin, ClaimFilter):
    pending_approval = None

    recipient = wb_filters.ModelChoiceFilter(
        label="Recipient",
        queryset=Entry.objects.all(),
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
        filter_params={"with_account": True},
        method="filter_recipient",
    )

    date = wb_filters.FinancialPerformanceDateRangeFilter(
        label="Date Range",
        method=lambda queryset, label, value: queryset,
        required=True,
        clearable=False,
        initial=current_financial_quarter,
    )

    class Meta:
        model = Claim
        fields = {"product": ["exact"], "account": ["exact"]}
        df_fields = {
            "unrealized_pnl__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Unrealized P&L", field_name="unrealized_pnl"
            ),
            "realized_pnl__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Realized P&L", field_name="realized_pnl"
            ),
            "total_pnl__lte": wb_filters.NumberFilter(lookup_expr="lte", label="Total P&L", field_name="total_pnl"),
            "total_invested__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Total Invested", field_name="total_invested"
            ),
            "performance__lte": wb_filters.NumberFilter(
                lookup_expr="lte", label="Performance", field_name="performance"
            ),
            "unrealized_pnl__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Unrealized P&L", field_name="unrealized_pnl"
            ),
            "realized_pnl__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Realized P&L", field_name="realized_pnl"
            ),
            "total_pnl__gte": wb_filters.NumberFilter(lookup_expr="gte", label="Total P&L", field_name="total_pnl"),
            "total_invested__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Total Invested", field_name="total_invested"
            ),
            "performance__gte": wb_filters.NumberFilter(
                lookup_expr="gte", label="Performance", field_name="performance"
            ),
        }


class CustomerClaimGroupByFilter(ClaimGroupByFilter):
    product = wb_filters.ModelChoiceFilter(
        label="Product",
        queryset=Product.objects.all(),
        endpoint=Product.get_representation_endpoint(),
        value_key=Product.get_representation_value_key(),
        label_key="{{title}} {{currency_repr}} - {{isin}} - {{bank_repr}}",
    )

    class Meta:
        model = Claim
        fields = {
            "account": ["exact"],
        }


class NegativeTermimalAccountPerProductFilterSet(wb_filters.FilterSet):
    sum_shares__gte = wb_filters.NumberFilter(lookup_expr="gte", field_name="sum_shares", label="Total Shares")
    sum_shares__lte = wb_filters.NumberFilter(lookup_expr="lte", field_name="sum_shares", label="Total Shares")

    only_approved = wb_filters.BooleanFilter(
        initial=False, label="Only Approved claims", method="filter_only_approve_claims"
    )

    def filter_only_approve_claims(self, queryset, name, value):
        if value is True:
            return queryset.filter(status=Claim.Status.APPROVED)
        return queryset

    class Meta:
        model = Claim
        fields = {
            "product": ["exact"],
            "account": ["exact"],
        }



from datetime import timedelta
from decimal import Decimal

from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbfdm.serializers.instruments.instruments import (
    InvestableUniverseRepresentationSerializer,
)

from wbportfolio.models import PortfolioRole, Trade
from wbportfolio.models.transactions.claim import Claim
from wbportfolio.serializers.custodians import CustodianRepresentationSerializer
from wbportfolio.serializers.registers import RegisterRepresentationSerializer

from .. import PortfolioRepresentationSerializer


class CopyClaimRepresentationSerializer(wb_serializers.RepresentationSerializer):
    claimant = wb_serializers.StringRelatedField()
    product = wb_serializers.StringRelatedField()
    account = wb_serializers.StringRelatedField()
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:claim-detail")

    class Meta:
        model = Claim
        fields = (
            "id",
            "shares",
            "claimant",
            "bank",
            "product",
            "account",
            "_detail",
        )


class TradeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:trade-detail")
    diff_shares = wb_serializers.DecimalField(max_digits=15, decimal_places=4, read_only=True)
    total_value = wb_serializers.DecimalField(max_digits=14, decimal_places=2, read_only=True, required=False)

    class Meta:
        model = Trade
        fields = (
            "id",
            "transaction_date",
            "total_value",
            "_detail",
            "claimed_shares",
            "diff_shares",
            "bank",
            "shares",
        )


class TradeClaimRepresentationSerializer(TradeRepresentationSerializer):
    def get_filter_params(self, request):
        claim_id = request.parser_context["view"].kwargs.get("pk", None)
        arg = {"marked_for_deletion": False}
        if claim_id:
            claim = Claim.objects.get(id=claim_id)
            arg.update(
                {
                    "underlying_instrument": claim.product.id,
                    "transaction_date__gte": claim.date - timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                    "transaction_date__lte": claim.date + timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                }
            )
            if claim.shares:
                if claim.shares > 0:
                    arg["shares__gte"] = claim.shares
                else:
                    arg["shares__lte"] = claim.shares
            return arg
        return {}


class TradeModelSerializer(wb_serializers.ModelSerializer):
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")
    _underlying_instrument = InvestableUniverseRepresentationSerializer(source="underlying_instrument")
    _currency = CurrencyRepresentationSerializer(source="currency")

    marked_for_deletion = wb_serializers.BooleanField(required=False, read_only=True)
    marked_as_internal = wb_serializers.BooleanField()

    approved_claimed_shares = wb_serializers.DecimalField(
        max_digits=10, decimal_places=2, required=False, read_only=True, default=Decimal(0.0)
    )
    pending_claimed_shares = wb_serializers.DecimalField(
        max_digits=10, decimal_places=2, required=False, read_only=True, default=Decimal(0.0)
    )
    completely_claimed = wb_serializers.BooleanField(required=False, read_only=True, default=Decimal(0.0))
    completely_claimed_if_approved = wb_serializers.BooleanField(required=False, read_only=True, default=Decimal(0.0))

    total_value = wb_serializers.DecimalField(max_digits=14, decimal_places=2, read_only=True, required=False)
    total_value_fx_portfolio = wb_serializers.DecimalField(
        max_digits=14, decimal_places=2, read_only=True, required=False
    )
    total_value_gross = wb_serializers.DecimalField(max_digits=14, decimal_places=2, read_only=True, required=False)
    total_value_gross_fx_portfolio = wb_serializers.DecimalField(
        max_digits=14, decimal_places=2, read_only=True, required=False
    )
    external_id = wb_serializers.CharField(required=False, read_only=True)
    value_date = wb_serializers.DateField(required=False, read_only=True)
    total_value_usd = wb_serializers.FloatField(default=0, read_only=True, label="Total Value ($)")
    total_value_gross_usd = wb_serializers.FloatField(default=0, read_only=True, label="Total Value Gross ($)")

    # We commented it because people needs to filter with lookup equals onto these fields
    # shares = wb_serializers.DecimalField(max_digits=16, decimal_places=2)
    # price = wb_serializers.DecimalField(max_digits=16, decimal_places=2)

    _register = RegisterRepresentationSerializer(source="register")
    claims = wb_serializers.PrimaryKeyRelatedField(queryset=Claim.objects.all(), label="Claims", many=True)
    _claims = CopyClaimRepresentationSerializer(source="claims", many=True)
    _custodian = CustodianRepresentationSerializer(source="custodian")
    _internal_trade = TradeRepresentationSerializer(
        source="internal_trade",
        filter_params={"only_internal_trade": True},
        optional_get_parameters={"underlying_instrument": "underlying_instrument", "transaction_date": "pivot_date"},
        depends_on=[{"field": "marked_as_internal", "options": {"activates_on": [True]}}],
    )

    @wb_serializers.register_only_instance_resource()
    def import_source_file_url(self, instance, request, user, *args, **kwargs):
        if PortfolioRole.is_portfolio_manager(user.profile, instance.underlying_instrument):
            if instance.import_source:
                return {"import_source": instance.import_source.file.url}
        return {}

    @wb_serializers.register_resource()
    def claims_list(self, instance, request, user):
        if instance.is_claimable:
            return {
                "claims": reverse(
                    "wbportfolio:trade-claim-list",
                    args=[instance.id],
                    request=request,
                )
            }
        return {}

    class Meta:
        model = Trade
        decorators = {
            "total_value": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "total_value_gross": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
            "total_value_usd": wb_serializers.decorator(decorator_type="text", position="left", value="{{$}}"),
            "total_value_gross_usd": wb_serializers.decorator(decorator_type="text", position="left", value="{{$}}"),
            "total_value_fx_portfolio": wb_serializers.decorator(
                position="left", value="{{_portfolio.currency_symbol}}"
            ),
            "total_value_gross_fx_portfolio": wb_serializers.decorator(
                position="left", value="{{_portfolio.currency_symbol}}"
            ),
            # "total_value_fx_portfolio": wb_serializers.decorator(decorator_type="text", position="left", value="{{_portfolio.currency_symbol}}}"),
            # "total_value_gross_fx_portfolio": wb_serializers.decorator(decorator_type="text", position="left", value="{{_portfolio.currency_symbol}}"),
        }
        read_only_fields = (
            "id",
            "shares",
            "price",
            "transaction_subtype",
            "approved_claimed_shares",
            "pending_claimed_shares",
            "claimed_shares",
            "completely_claimed",
            "completely_claimed_if_approved",
            "register",
            "_register",
            "pending",
            "bank",
            "marked_for_deletion",
            "_claims",
            "claims",
            "portfolio",
            "_portfolio",
            "underlying_instrument",
            "_underlying_instrument",
            "transaction_date",
            "book_date",
            "value_date",
            "currency",
            "_currency",
            "currency_fx_rate",
            "total_value",
            "total_value_fx_portfolio",
            "total_value_gross",
            "total_value_gross_fx_portfolio",
            "external_id",
            "total_value_usd",
            "total_value_gross_usd",
            "_additional_resources",
        )

        fields = (
            "id",
            "shares",
            "price",
            "transaction_subtype",
            "approved_claimed_shares",
            "pending_claimed_shares",
            "claimed_shares",
            "completely_claimed",
            "completely_claimed_if_approved",
            "register",
            "_register",
            "pending",
            "bank",
            "marked_for_deletion",
            "_claims",
            "claims",
            "portfolio",
            "_portfolio",
            "underlying_instrument",
            "_underlying_instrument",
            "transaction_date",
            "book_date",
            "value_date",
            "currency",
            "_currency",
            "currency_fx_rate",
            "total_value",
            "total_value_fx_portfolio",
            "total_value_gross",
            "total_value_gross_fx_portfolio",
            "external_id",
            "comment",
            "total_value_usd",
            "total_value_gross_usd",
            "_custodian",
            "custodian",
            "marked_as_internal",
            "internal_trade",
            "_internal_trade",
            "_additional_resources",
        )



from django.db import models
from wbcore.contrib.io.mixins import ImportMixin

from wbportfolio.import_export.handlers.dividend import DividendImportHandler

from .transactions import TransactionMixin


class DividendTransaction(TransactionMixin, ImportMixin, models.Model):
    import_export_handler_class = DividendImportHandler

    class DistributionMethod(models.TextChoices):
        PAYMENT = "Payment", "Payment"
        REINVESTMENT = "Reinvestment", "Reinvestment"

    ex_date = models.DateField(
        verbose_name="Ex-Dividend Date",
        help_text="The date on which the stock starts trading without the dividend",
    )
    record_date = models.DateField(
        verbose_name="Record Date",
        help_text="The date on which the holder must own the shares to be eligible for the dividend",
    )
    distribution_method = models.CharField(
        max_length=255, verbose_name="Type", choices=DistributionMethod.choices, default=DistributionMethod.PAYMENT
    )
    retrocession = models.FloatField(default=1)
    price = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        help_text="The amount paid per share",
        verbose_name="DPS",
    )
    total_value_gross = models.GeneratedField(
        expression=models.F("price") * models.F("shares") * models.F("retrocession"),
        output_field=models.DecimalField(
            max_digits=20,
            decimal_places=4,
        ),
        db_persist=True,
    )

    def save(self, *args, **kwargs):
        self.pre_save()
        if not self.record_date and self.ex_date:
            self.record_date = self.ex_date
        elif self.record_date and not self.ex_date:
            self.ex_date = self.record_date
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.total_value} - {self.value_date:%d.%m.%Y} : {str(self.underlying_instrument)} (in {str(self.portfolio)})"



# Generated by Django 5.0.12 on 2025-09-15 13:45

import django.core.validators
from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0088_orderproposal_total_effective_portfolio_contribution'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderproposal',
            name='min_weighting',
            field=models.DecimalField(decimal_places=8, default=Decimal('0'), help_text='The minimum weight allowed for this order proposal ', max_digits=9, verbose_name='Minimum Weight'),
        ),
        migrations.AddField(
            model_name='portfolio',
            name='default_order_proposal_min_order_value',
            field=models.IntegerField(default=0, verbose_name='Default Order Proposal Minimum Order Value'),
        ),
        migrations.AddField(
            model_name='portfolio',
            name='default_order_proposal_min_weighting',
            field=models.DecimalField(decimal_places=8, default=Decimal('0'), max_digits=9,
                                      verbose_name='Default Order Proposal Minimum Weight'),
        ),
        migrations.AddField(
            model_name='portfolio',
            name='default_order_proposal_total_cash_weight',
            field=models.DecimalField(decimal_places=4, default=Decimal('0'), max_digits=5,
                                      verbose_name='Default Order Proposal Total Cash Weight'),
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='min_weighting',
            field=models.DecimalField(decimal_places=8, default=Decimal('0'),
                                      help_text='The minimum weight allowed for this order proposal ', max_digits=9,
                                      validators=[django.core.validators.MinValueValidator(Decimal('0')),
                                                  django.core.validators.MaxValueValidator(Decimal('1'))],
                                      verbose_name='Minimum Weight'),
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='total_cash_weight',
            field=models.DecimalField(decimal_places=4, default=Decimal('0'),
                                      help_text='The desired percentage for the cash component. The remaining percentage (100% minus this value) will be allocated to total target weighting. Default is 0%.',
                                      max_digits=5, validators=[django.core.validators.MinValueValidator(Decimal('0')),
                                                                django.core.validators.MaxValueValidator(Decimal('1'))],
                                      verbose_name='Total Cash Weight'),
        ),
        migrations.AlterField(
            model_name='portfolio',
            name='default_order_proposal_min_weighting',
            field=models.DecimalField(decimal_places=8, default=Decimal('0'), max_digits=9,
                                      validators=[django.core.validators.MinValueValidator(Decimal('0')),
                                                  django.core.validators.MaxValueValidator(Decimal('1'))],
                                      verbose_name='Default Order Proposal Minimum Weight'),
        ),
        migrations.AlterField(
            model_name='portfolio',
            name='default_order_proposal_total_cash_weight',
            field=models.DecimalField(decimal_places=4, default=Decimal('0'), max_digits=5,
                                      validators=[django.core.validators.MinValueValidator(Decimal('0')),
                                                  django.core.validators.MaxValueValidator(Decimal('1'))],
                                      verbose_name='Default Order Proposal Total Cash Weight'),
        ),
    ]



from wbfdm.contrib.metric.dto import MetricField, MetricKey

PORTFOLIO_EBIT = MetricKey(
    key="portfolio_ebit",
    label="EBIT",
    subfields=[
        MetricField(
            key="fy_3",
            label="FY-3",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy_2",
            label="FY-2",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy_1",
            label="FY-1",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy0",
            label="FY0",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(key="fy1", label="FY1", decorators=[{"position": "left", "value": "$"}]),
        MetricField(
            key="fy2",
            label="FY2",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy3",
            label="FY3",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy4",
            label="FY4",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
    ],
)

PORTFOLIO_TOTAL_ASSETS = MetricKey(
    key="portfolio_total_assets",
    label="Total Assets",
    subfields=[
        MetricField(
            key="fy_3",
            label="FY-3",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy_2",
            label="FY-2",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy_1",
            label="FY-1",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy0",
            label="FY0",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(key="fy1", label="FY1", decorators=[{"position": "left", "value": "$"}]),
        MetricField(
            key="fy2",
            label="FY2",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy3",
            label="FY3",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy4",
            label="FY4",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
    ],
)

PORTFOLIO_LIABILITIES = MetricKey(
    key="portfolio_liabilities",
    label="Liabilities",
    subfields=[
        MetricField(
            key="fy_3",
            label="FY-3",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy_2",
            label="FY-2",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy_1",
            label="FY-1",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy0",
            label="FY0",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(key="fy1", label="FY1", decorators=[{"position": "left", "value": "$"}]),
        MetricField(
            key="fy2",
            label="FY2",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy3",
            label="FY3",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy4",
            label="FY4",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
    ],
)

PORTFOLIO_CAPITAL_EMPLOYED = MetricKey(
    key="portfolio_capital_employed",
    label="Capital Employed",
    subfields=[
        MetricField(
            key="fy_3",
            label="FY-3",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy_2",
            label="FY-2",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy_1",
            label="FY-1",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy0",
            label="FY0",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(key="fy1", label="FY1", decorators=[{"position": "left", "value": "$"}]),
        MetricField(
            key="fy2",
            label="FY2",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy3",
            label="FY3",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
        MetricField(
            key="fy4",
            label="FY4",
            list_display_kwargs={"show": "open"},
            decorators=[{"position": "left", "value": "$"}],
        ),
    ],
)

PORTFOLIO_ROCE = MetricKey(
    key="portfolio_roce",
    label="ROCE",
    subfields=[
        MetricField(
            key="fy_3", label="FY-3", list_display_kwargs={"show": "open"}, serializer_kwargs={"percent": True}
        ),
        MetricField(
            key="fy_2", label="FY-2", list_display_kwargs={"show": "open"}, serializer_kwargs={"percent": True}
        ),
        MetricField(
            key="fy_1", label="FY-1", list_display_kwargs={"show": "open"}, serializer_kwargs={"percent": True}
        ),
        MetricField(key="fy0", label="FY0", list_display_kwargs={"show": "open"}, serializer_kwargs={"percent": True}),
        MetricField(key="fy1", label="FY1", serializer_kwargs={"percent": True}),
        MetricField(key="fy2", label="FY2", list_display_kwargs={"show": "open"}, serializer_kwargs={"percent": True}),
        MetricField(key="fy3", label="FY3", list_display_kwargs={"show": "open"}, serializer_kwargs={"percent": True}),
        MetricField(key="fy4", label="FY4", list_display_kwargs={"show": "open"}, serializer_kwargs={"percent": True}),
    ],
)



from wbcore.metadata.configs.titles import TitleViewConfig


class ESGMetricAggregationPortfolioPandasTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if esg_aggregation := self.view.esg_aggregation:
            metric_label = f"{esg_aggregation.value} ({esg_aggregation.get_aggregation().value})"
        else:
            metric_label = "ESG metric"
        return f"{self.view.portfolio}: {metric_label}"



from typing import Generator

from wbcompliance.models.risk_management import backend
from wbcompliance.models.risk_management.dispatch import register
from wbcore import serializers as wb_serializers
from wbfdm.analysis.esg.enums import ESGAggregation
from wbfdm.analysis.esg.esg_analysis import DataLoader
from wbfdm.models import Instrument

from wbportfolio.pms.typing import Portfolio as PortfolioDTO

from .mixins import ActivePortfolioRelationshipMixin


@register("ESG Aggregation Portfolio Rule Backend", rule_group_key="portfolio")
class RuleBackend(ActivePortfolioRelationshipMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.esg_aggregation = ESGAggregation[self.esg_aggregation]

    @classmethod
    def get_serializer_class(cls) -> wb_serializers.Serializer:
        class RuleBackendSerializer(wb_serializers.Serializer):
            esg_aggregation = wb_serializers.ChoiceField(
                choices=ESGAggregation.choices(),
                default=ESGAggregation.GHG_EMISSIONS_SCOPE_1.name,
            )

            @classmethod
            def get_parameter_fields(cls):
                return [
                    "esg_aggregation",
                ]

        return RuleBackendSerializer

    def _process_dto(self, portfolio: PortfolioDTO, **kwargs) -> Generator[backend.IncidentResult, None, None]:
        esg_data = self.esg_aggregation.get_esg_data(Instrument.objects.filter(id__in=portfolio.positions_map.keys()))
        df = portfolio.to_df(exclude_cash=True)
        df["total_value"] = (df["price"] * df["shares"] * df["currency_fx_rate"]).astype(float)
        df = df[["total_value", "weighting", "underlying_instrument"]].set_index("underlying_instrument")
        df["weighting"] = df["weighting"] / df["weighting"].sum()
        dataloader = DataLoader(
            df["weighting"].astype(float), esg_data, self.evaluation_date, total_value_fx_usd=df["total_value"]
        )
        metrics = dataloader.compute(self.esg_aggregation)
        for threshold in self.thresholds:
            numerical_range = threshold.numerical_range
            incident_df = metrics[(metrics >= numerical_range[0]) & (metrics < numerical_range[1])]
            for instrument_id, metric in incident_df.to_dict().items():
                instrument = Instrument.objects.get(id=instrument_id)
                breached_value = metric

                if metric < 0:
                    breached_value = f'<span style="color:red">{breached_value}</span>'
                else:
                    breached_value = f'<span style="color:green">{breached_value}</span>'
                yield backend.IncidentResult(
                    breached_object=instrument,
                    breached_object_repr=str(instrument),
                    breached_value=breached_value,
                    report_details={"Aggregation": self.esg_aggregation.value},
                    severity=threshold.severity,
                )



import factory

from wbportfolio.models import Rebalancer, RebalancingModel


class RebalancingModelFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = RebalancingModel

    name = factory.Faker("name")
    class_path = "wbportfolio.rebalancing.models.equally_weighted.EquallyWeightedRebalancing"


class RebalancerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Rebalancer

    portfolio = factory.SubFactory("wbportfolio.factories.portfolios.PortfolioFactory")
    rebalancing_model = factory.SubFactory(RebalancingModelFactory)
    parameters = dict()
    apply_order_proposal_automatically = False
    frequency = "RRULE:FREQ=MONTHLY;"
    activation_date = None



from unittest.mock import MagicMock, PropertyMock, patch

import pytest
from django.conf import settings

from wbportfolio.order_routing import ExecutionStatus
from wbportfolio.order_routing.router import Router
from wbportfolio.pms.typing import Order


@pytest.fixture
def mock_adapter():
    adapter = MagicMock()
    return adapter


@pytest.fixture
def router(mock_adapter):
    return Router(adapter=mock_adapter)


def test_submit_as_draft_from_settings(monkeypatch, router):
    # Test default True if settings attribute missing
    monkeypatch.setattr(settings, "ORDER_ROUTING_AS_DRAFT", True)
    monkeypatch.setattr(settings, "DEBUG", False)
    assert router.submit_as_draft is True

    monkeypatch.setattr(settings, "ORDER_ROUTING_AS_DRAFT", False)
    monkeypatch.setattr(settings, "DEBUG", True)
    assert router.submit_as_draft is True

    monkeypatch.setattr(settings, "ORDER_ROUTING_AS_DRAFT", True)
    monkeypatch.setattr(settings, "DEBUG", True)
    assert router.submit_as_draft is True

    monkeypatch.setattr(settings, "ORDER_ROUTING_AS_DRAFT", False)
    monkeypatch.setattr(settings, "DEBUG", False)
    assert router.submit_as_draft is False


@patch.object(Router, "submit_as_draft", new_callable=PropertyMock)
def test_submit_rebalancing_calls_adapter_as_draft(mock_property, router, mock_adapter):
    mock_property.return_value = True
    orders = [MagicMock(spec=Order), MagicMock(spec=Order)]
    serialized_orders = ["serialized_order1", "serialized_order2"]  # simplified serialized orders as items
    confirmed_items = ["confirmed_order1", "confirmed_order2"]  # simplified deserialized orders from items
    msg = "Success message"

    mock_adapter.serialize_orders.return_value = serialized_orders
    mock_adapter.submit_rebalancing.return_value = (confirmed_items, msg)
    mock_adapter.deserialize_items.return_value = orders

    result_orders, (status, message) = router.submit_rebalancing(orders)
    assert result_orders == orders
    assert status == ExecutionStatus.IN_DRAFT
    assert message == msg
    mock_adapter.serialize_orders.assert_called_once_with(orders)
    mock_adapter.submit_rebalancing.assert_called_once_with(serialized_orders, as_draft=True)
    mock_adapter.deserialize_items.assert_called_once_with(confirmed_items)


@patch.object(Router, "submit_as_draft", new_callable=PropertyMock)
def test_submit_rebalancing_calls_adapter(mock_property, router, mock_adapter):
    mock_property.return_value = False
    orders = [MagicMock(spec=Order), MagicMock(spec=Order)]
    serialized_orders = ["serialized_order1", "serialized_order2"]  # simplified serialized orders as items
    confirmed_items = ["confirmed_order1", "confirmed_order2"]  # simplified deserialized orders from items
    msg = "Success message"

    mock_adapter.serialize_orders.return_value = serialized_orders
    mock_adapter.submit_rebalancing.return_value = (confirmed_items, msg)
    mock_adapter.deserialize_items.return_value = orders

    result_orders, (status, message) = router.submit_rebalancing(orders)
    assert result_orders == orders
    assert status == ExecutionStatus.PENDING
    assert message == msg
    mock_adapter.serialize_orders.assert_called_once_with(orders)
    mock_adapter.submit_rebalancing.assert_called_once_with(serialized_orders, as_draft=False)
    mock_adapter.deserialize_items.assert_called_once_with(confirmed_items)


def test_get_rebalance_status_returns_adapter_status(router, mock_adapter):
    expected_status = ExecutionStatus.PENDING
    expected_msg = "Status message"
    mock_adapter.get_rebalance_status.return_value = (expected_status, expected_msg)

    status, msg = router.get_rebalance_status()
    assert status == expected_status
    assert msg == expected_msg
    mock_adapter.get_rebalance_status.assert_called_once()


def test_cancel_rebalancing_returns_adapter_result(router, mock_adapter):
    mock_adapter.cancel_current_rebalancing.return_value = True
    result = router.cancel_rebalancing()
    assert result is True
    mock_adapter.cancel_current_rebalancing.assert_called_once()


def test_get_current_rebalancing_request_returns_deserialized(router, mock_adapter):
    serialized_orders = ["order1", "order2"]
    deserialized_orders = [MagicMock(spec=Order), MagicMock(spec=Order)]
    mock_adapter.get_current_rebalancing.return_value = serialized_orders
    mock_adapter.deserialize_items.return_value = deserialized_orders

    result = router.get_current_rebalancing_request()
    assert result == deserialized_orders
    mock_adapter.get_current_rebalancing.assert_called_once()
    mock_adapter.deserialize_items.assert_called_once_with(serialized_orders)



from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.models import Company
from wbcore.contrib.directory.serializers import CompanyRepresentationSerializer
from wbcore.contrib.geography.serializers import GeographyRepresentationSerializer
from wbfdm.serializers.instruments import InstrumentModelSerializer

from wbportfolio.models import ProductGroup, ProductGroupRepresentant
from wbportfolio.serializers.portfolios import PortfolioRepresentationSerializer


class ProductGroupRepresentantRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _representant = CompanyRepresentationSerializer(source="representant")
    _country = GeographyRepresentationSerializer(source="country", filter_params={"level": 1})

    class Meta:
        model = ProductGroupRepresentant
        fields = (
            "id",
            "product_group",
            "representant",
            "_representant",
            "country",
            "_country",
        )


class ProductGroupRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:product_group-detail")

    class Meta:
        model = ProductGroup
        fields = ("id", "computed_str", "name_repr", "_detail")


class ProductGroupModelSerializer(InstrumentModelSerializer):
    id_repr = wb_serializers.CharField(source="id", read_only=True, label="ID")
    _management_company = CompanyRepresentationSerializer(source="management_company")
    management_company = wb_serializers.PrimaryKeyRelatedField(
        queryset=Company.objects.all().select_related("entry_ptr")
    )
    _representants = ProductGroupRepresentationSerializer(source="representants", many=True)

    _depositary = CompanyRepresentationSerializer(source="depositary")
    depositary = wb_serializers.PrimaryKeyRelatedField(queryset=Company.objects.all().select_related("entry_ptr"))

    _transfer_agent = CompanyRepresentationSerializer(source="transfer_agent")
    transfer_agent = wb_serializers.PrimaryKeyRelatedField(queryset=Company.objects.all().select_related("entry_ptr"))

    _administrator = CompanyRepresentationSerializer(source="administrator")
    administrator = wb_serializers.PrimaryKeyRelatedField(queryset=Company.objects.all().select_related("entry_ptr"))

    _investment_manager = CompanyRepresentationSerializer(source="investment_manager")
    investment_manager = wb_serializers.PrimaryKeyRelatedField(
        queryset=Company.objects.all().select_related("entry_ptr")
    )

    _auditor = CompanyRepresentationSerializer(source="auditor")
    auditor = wb_serializers.PrimaryKeyRelatedField(queryset=Company.objects.all().select_related("entry_ptr"))

    _paying_agent = CompanyRepresentationSerializer(source="paying_agent")
    paying_agent = wb_serializers.PrimaryKeyRelatedField(queryset=Company.objects.all().select_related("entry_ptr"))

    _portfolios = PortfolioRepresentationSerializer(source="portfolios", many=True)

    class Meta(InstrumentModelSerializer.Meta):
        model = ProductGroup
        fields = (
            "id_repr",
            "type",
            "category",
            "umbrella",
            "management_company",
            "_management_company",
            "depositary",
            "_depositary",
            "transfer_agent",
            "_transfer_agent",
            "administrator",
            "_administrator",
            "investment_manager",
            "_investment_manager",
            "auditor",
            "_auditor",
            "paying_agent",
            "_paying_agent",
            "representants",
            "_representants",
            "_portfolios",
            "portfolios",
        ) + InstrumentModelSerializer.Meta.fields



from decimal import Decimal
from typing import Any, Dict

from django.db import models
from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler, ImportState
from wbcore.contrib.io.utils import nest_row
from wbfdm.import_export.handlers.instrument import InstrumentImportHandler

from wbportfolio.pms.typing import Portfolio, Position


class OrderImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbportfolio.Order"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)
        self.order_proposal = None

    def process_object(
        self,
        data: Dict[str, Any],
        **kwargs,
    ):
        from wbportfolio.models import OrderProposal

        data = nest_row(data)
        underlying_instrument = self.instrument_handler.process_object(
            data["underlying_instrument"], only_security=False, read_only=True
        )[0]
        self.order_proposal = OrderProposal.objects.get(id=data.pop("order_proposal_id"))
        weighting = data.get("target_weight", data.get("weighting"))
        shares = data.get("target_shares", data.get("shares", 0))
        if weighting is None:
            raise DeserializationError("We couldn't figure out the target weight column")
        position_dto = Position(
            underlying_instrument=underlying_instrument.id,
            instrument_type=underlying_instrument.instrument_type.id,
            weighting=Decimal(weighting),
            shares=Decimal(shares),
            currency=underlying_instrument.currency,
            date=self.order_proposal.trade_date,
            is_cash=underlying_instrument.is_cash,
        )
        return position_dto, ImportState.CREATED

    def _get_history(self, history: Dict[str, Any]) -> models.QuerySet:
        from wbportfolio.models.orders.order_proposals import OrderProposal

        if order_proposal_id := history.get("order_proposal_id"):
            # if a order proposal is provided, we delete the existing history first as otherwise, it would mess with the target weight computation
            order_proposal = OrderProposal.objects.get(id=order_proposal_id)
            order_proposal.orders.all().delete()
        return self.model.objects.none()

    def _post_processing_objects(self, positions: list[Position], *args, **kwargs):
        total_weight = sum(map(lambda p: p.weighting, positions))
        if cash_weight := Decimal("1") - total_weight:
            cash_component = self.order_proposal.cash_component
            positions.append(
                Position(
                    underlying_instrument=cash_component.id,
                    instrument_type=cash_component.instrument_type.id,
                    weighting=cash_weight,
                    currency=cash_component.currency,
                    date=self.order_proposal.trade_date,
                    is_cash=cash_component.is_cash,
                )
            )
        self.order_proposal.reset_orders(target_portfolio=Portfolio(positions), ignore_existing_cash=True)



from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class RegisterButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {bt.WidgetButton(key="trades", label="Trades", icon=WBIcon.TRADE.icon)}

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



from typing import TYPE_CHECKING

import pandas as pd
from pandas.tseries.offsets import BDay
from schwifty import IBAN

from wbportfolio.models.portfolio import Portfolio

if TYPE_CHECKING:
    from wbcore.contrib.io.models import ImportSource

# Bank specific Information for Societe Generale Luxembourg
BANK_COUNTRY_CODE = "LU"
BANK_CODE = "060"


def get_underlying_instrument(row: pd.Series) -> dict[str, str]:
    """Get the underlying instrument from the row"""
    return {
        "instrument_type": "equity",
        "isin": str(row["isin"]),
        "name": str(row["name"]),
    }


def get_portfolio_id(row: pd.Series) -> int:
    """Get the portfolio id from the bank account number

    Raises: Portfolio.DoesNotExist: We raise an error intentionally if the portfolio does not exist to make the import fail
    """
    iban = str(IBAN.generate(BANK_COUNTRY_CODE, bank_code=BANK_CODE, account_code=str(row["account_number"])))
    return Portfolio.all_objects.get(bank_accounts__iban=iban).id


def parse(import_source: "ImportSource") -> dict:
    """Parse the custodian positions file into a dictionary"""

    # Get df from csv and rename columns and group by account number, isin, and date to sum the shares per date, account, and isin
    df = pd.read_csv(import_source.file.open(), encoding="latin1").rename(  # type: ignore
        columns={
            "Account number": "account_number",
            "ISIN code": "isin",
            "Quantity": "initial_shares",
            "Position date": "date",
            "Security name": "name",
        }
    )

    df = pd.DataFrame(
        df[["account_number", "isin", "initial_shares", "date", "name"]]
        .groupby(["account_number", "isin", "date", "name"])
        .sum()
    )

    # add weight per account number
    df["weighting"] = df.groupby(["account_number", "date"])["initial_shares"].transform(lambda x: x / x.sum())

    # reset index to remove the multi index from the groupby
    df = df.reset_index()

    # parse datetime string to date string and set to 1 day earlier
    df["date"] = pd.to_datetime(df["date"], format="%m/%d/%Y %I:%M:%S %p") - BDay(1)
    df["date"] = df["date"].dt.strftime("%Y-%m-%d")

    df["underlying_instrument"] = df.apply(get_underlying_instrument, axis=1)
    df["portfolio"] = df.apply(get_portfolio_id, axis=1)
    df["estimated"] = True

    # remove name column
    df = df.drop(columns=["name", "account_number"])
    return {"data": df.to_dict("records")}



# Generated by Django 5.0.14 on 2025-07-02 13:10

from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0077_remove_transaction_currency_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='trade',
            name='drift_factor',
            field=models.DecimalField(decimal_places=6, default=Decimal('1'), help_text='Drift factor to be applied to the previous portfolio weight to get the actual effective weight including daily return', max_digits=16, verbose_name='Drift Factor'),
        ),
        migrations.AlterField(
            model_name='trade',
            name='weighting',
            field=models.DecimalField(decimal_places=8, default=Decimal('0'),
                                      help_text='The weight to be multiplied against the target', max_digits=9,
                                      verbose_name='Weight'),
        ),
    ]



import numpy as np
import pandas as pd


def parse(import_source):
    data = []
    if (
        (data_backend := import_source.source.data_backend)
        and (backend_class := data_backend.backend_class)
        and (default_mapping := backend_class.DEFAULT_MAPPING)
    ):
        df = pd.read_json(import_source.file, orient="records")
        if not df.empty:
            if date_label := getattr(backend_class, "DATE_LABEL", None):
                df[date_label] = pd.to_datetime(df[date_label])
                df = df.sort_values(by=date_label, ascending=True)
                if reindex_method := getattr(backend_class, "REINDEX_METHOD", None):
                    df = df.set_index(date_label)
                    timeline = pd.date_range(start=df.index.min(), end=df.index.max(), freq="B")
                    df = df.reindex(timeline, method=reindex_method)
                    df = df.reset_index(names=date_label)
                df[date_label] = df[date_label].dt.strftime("%Y-%m-%d")
        df = df.replace([np.inf, -np.inf, np.nan], None).rename(columns=default_mapping)
        df = df.drop(columns=df.columns.difference([*default_mapping.values()]))

        if none_nullable_fields := getattr(backend_class, "NONE_NULLABLE_FIELDS", None):
            df = df.dropna(subset=none_nullable_fields, how="any")

        data = df.to_dict("records")
    return {"data": data}



from decimal import Decimal

import pandas as pd
from django.db import models
from django.db.models import F, Q, Sum
from django.dispatch import receiver
from wbcore.models import WBModel
from wbcore.signals import pre_merge
from wbfdm.models import InstrumentType
from wbfdm.models.instruments.instrument_prices import InstrumentPrice

from wbportfolio.models.products import FeeProductPercentage, Product

from .mixins.instruments import PMSInstrumentAbstractModel


class ProductGroup(PMSInstrumentAbstractModel):
    class ProductGroupType(models.TextChoices):
        FUND = "Fund"

    class ProductGroupCategory(models.TextChoices):
        UCITS = "UCITS", "UCITS"
        OTHER_FUNDS = "OTHER_FUNDS", "Other funds for traditional investments"

    type = models.CharField(
        max_length=64,
        verbose_name="Type",
        choices=ProductGroupType.choices,
        default=ProductGroupType.FUND,
    )
    category = models.CharField(
        max_length=64,
        choices=ProductGroupCategory.choices,
        default=ProductGroupCategory.UCITS,
    )

    umbrella = models.CharField(max_length=255, null=True, blank=True)
    management_company = models.ForeignKey(
        "directory.Company",
        related_name="management_company_product_groups",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    depositary = models.ForeignKey(
        "directory.Company",
        related_name="depositary_product_groups",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    transfer_agent = models.ForeignKey(
        "directory.Company",
        related_name="transfer_agent_product_groups",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    administrator = models.ForeignKey(
        "directory.Company",
        related_name="administrator_groups",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    investment_manager = models.ForeignKey(
        "directory.Company",
        related_name="investment_manager_groups",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    auditor = models.ForeignKey(
        "directory.Company",
        related_name="auditor_groups",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    paying_agent = models.ForeignKey(
        "directory.Company",
        related_name="paying_agent_groups",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )

    @property
    def products(self):  # for backward compatibility
        return Product.objects.filter(id__in=self.children.values("id"))

    def save(self, *args, **kwargs):
        self.instrument_type = InstrumentType.objects.get_or_create(
            key="product_group", defaults={"name": "Product Group", "short_name": "Product Group"}
        )[0]
        self.is_managed = True
        super().save(*args, **kwargs)

    def active_products(self, val_date):
        return self.products.filter(
            Q(inception_date__isnull=False)
            & Q(inception_date__lte=val_date)
            & (Q(delisted_date__isnull=True) | Q(delisted_date__gt=val_date))
        )

    def compute_str(self):
        return f"{self.name} ({self.umbrella})"

    class Meta:
        verbose_name = "Product Group"
        verbose_name_plural = "Product Groups"

    def get_fund_product_table(self, val_date):
        products = self.active_products(val_date).annotate(
            current_net_value=InstrumentPrice.subquery_closest_value("net_value", val_date, instrument_pk_name="pk"),
        )
        df = pd.DataFrame(
            products.values(
                "name",
                "isin",
                "ticker",
                "refinitiv_identifier_code",
                "currency__key",
                "currency__symbol",
                "dividend",
                "minimum_subscription",
                "inception_date",
                "current_net_value",
            )
        )

        if not df.empty:
            df["total_performance_fee"] = df["isin"].apply(
                lambda x: Product.objects.get(isin=x).get_fees_percent(
                    val_date, fee_type=FeeProductPercentage.Type.PERFORMANCE, net=False
                )
            )
            df["management_fees"] = df["isin"].apply(
                lambda x: Product.objects.get(isin=x).get_fees_percent(
                    val_date, fee_type=FeeProductPercentage.Type.MANAGEMENT
                )
            )
            df.management_fees = df.management_fees.apply(lambda x: f"{x:,.2%}" if x is not None else "")
            df.total_performance_fee = df.total_performance_fee.apply(lambda x: f"{x:,.2%}" if x is not None else "")
            df.current_net_value = df.current_net_value.apply(lambda x: f"{x:,.2f}" if x is not None else "")
            df.inception_date = df.inception_date.apply(lambda x: f"{x:%Y-%m-%d}" if x is not None else "")

            df = df[
                [
                    "name",
                    "isin",
                    "ticker",
                    "refinitiv_identifier_code",
                    "currency__key",
                    "currency__symbol",
                    "dividend",
                    "management_fees",
                    "total_performance_fee",
                    "minimum_subscription",
                    "current_net_value",
                    "inception_date",
                ]
            ]
            df = df.rename(
                columns={
                    "name": "share class",
                    "ticker": "bloomberg",
                    "currency__key": "currency",
                    "currency__symbol": "currency symbol",
                    "minimum_subscription": "min subscription",
                    "management_fees": "mgmt. fees",
                    "total_performance_fee": "perf. fees",
                    "inception_date": "launch date",
                    "refinitiv_identifier_code": "reuters",
                    "current_net_value": "price",
                }
            )

            return df

    def get_total_fund_aum(self, val_date=None):
        return Product.annotate_last_aum(
            Product.objects.filter(parent=self, is_invested=True), val_date=val_date
        ).aggregate(s=Sum(F("assets_under_management_usd")))["s"] or Decimal(0)

    def total_shares(self, val_date):
        return sum([product.total_shares(val_date) for product in self.products.all()])

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name_repr}}"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:product_grouprepresentation-list"


class ProductGroupRepresentant(WBModel):
    product_group = models.ForeignKey("ProductGroup", related_name="representants", on_delete=models.CASCADE)
    representant = models.ForeignKey("directory.Company", on_delete=models.PROTECT)
    country = models.ForeignKey(
        "geography.Geography", on_delete=models.PROTECT, limit_choices_to={"level": 1}, null=True, blank=True
    )

    def __str__(self):
        return f"{self.product_group.identifier} {self.representant.name} ({self.country.name})"

    class Meta:
        verbose_name = "Product Group Representant"
        verbose_name_plural = "Product Group Representants"
        constraints = [
            models.UniqueConstraint(fields=("product_group", "country"), name="unique_country_product_group")
        ]

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{representant}} {{country}}"

    @classmethod
    def get_representation_endpoint(cls):
        return None

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:product_group"


@receiver(pre_merge, sender="directory.Company")
def pre_merge_company_productgroup(sender: models.Model, merged_object, main_object, **kwargs):
    ProductGroupRepresentant.objects.filter(representant=merged_object).update(representant=main_object)
    ProductGroup.objects.filter(management_company=merged_object).update(management_company=main_object)
    ProductGroup.objects.filter(depositary=merged_object).update(depositary=main_object)
    ProductGroup.objects.filter(transfer_agent=merged_object).update(transfer_agent=main_object)
    ProductGroup.objects.filter(administrator=merged_object).update(administrator=main_object)
    ProductGroup.objects.filter(investment_manager=merged_object).update(investment_manager=main_object)
    ProductGroup.objects.filter(auditor=merged_object).update(auditor=main_object)
    ProductGroup.objects.filter(paying_agent=merged_object).update(paying_agent=main_object)



from wbcore.menus import ItemPermission, MenuItem

from wbportfolio.permissions import is_portfolio_manager

ADJUSTMENT_MENUITEM = MenuItem(
    label="Adjustment",
    endpoint="wbportfolio:adjustment-list",
    permission=ItemPermission(permissions=["wbportfolio.view_adjustment"], method=is_portfolio_manager),
)



from datetime import date

import pytest
from django.contrib.auth.models import Permission
from django.db import models
from django.test.client import RequestFactory
from faker import Faker
from psycopg.types.range import DateRange
from rest_framework.reverse import reverse
from rest_framework.test import force_authenticate
from wbfdm.models import InstrumentPrice

from wbportfolio.models import InstrumentPortfolioThroughModel, Product
from wbportfolio.viewsets import ProductModelViewSet

fake = Faker()


@pytest.mark.django_db
class TestProductModelViewSet:
    def create_queryset(self, portfolio_factory, product_factory, instrument_price_factory, customer_trade_factory):
        for portfolio, product in zip(
            portfolio_factory.create_batch(4, invested_timespan=DateRange(date.min, date.max))
            + portfolio_factory.create_batch(2, invested_timespan=DateRange(date.min, date.max)),
            product_factory.create_batch(6),
            strict=False,
        ):
            InstrumentPortfolioThroughModel.objects.update_or_create(
                instrument=product, defaults={"portfolio": portfolio}
            )
            prices = instrument_price_factory.create_batch(10, instrument=product, calculated=False)
            for price in prices:
                trade = customer_trade_factory.create(
                    transaction_date=price.date, underlying_instrument=product, portfolio=product.primary_portfolio
                )
                price.outstanding_shares = trade.shares
                price.save()

    def test_list(self, superuser):
        url = reverse("wbportfolio:product-list")
        request = RequestFactory().get(url)
        force_authenticate(request, user=superuser)
        view = ProductModelViewSet.as_view({"get": "list"})
        response = view(request)
        assert response.status_code == 200

    def test_get_aggregates(
        self, superuser, portfolio_factory, product_factory, instrument_price_factory, customer_trade_factory
    ):
        url = reverse("wbportfolio:product-list")
        self.create_queryset(portfolio_factory, product_factory, instrument_price_factory, customer_trade_factory)
        superuser.is_internal = True
        superuser.save()
        request = RequestFactory().get(f"{url}?price_date={InstrumentPrice.objects.latest('date').date}")
        request.user = superuser
        force_authenticate(request, user=superuser)

        view = ProductModelViewSet(request=request)
        view.request = request
        queryset = view.get_queryset()
        agg = view.get_aggregates(queryset, queryset)
        aum = float(
            queryset.filter(is_invested=True).aggregate(s=models.Sum("assets_under_management_usd"))["s"] or 0.0
        )
        assert float(list(agg["assets_under_management_usd"].values())[0]) == pytest.approx(aum, rel=1e-4)

    def test_queryset_superuser(self, superuser, product_factory):
        url = reverse("wbportfolio:product-list")
        request = RequestFactory().get(url)
        force_authenticate(request, user=superuser)
        view = ProductModelViewSet.as_view({"get": "list"})
        product_factory.create_batch(5)
        response = view(request)
        assert response.status_code == 200
        ids = [r["id"] for r in response.data["results"]]
        assert set(ids) == set(list(Product.objects.values_list("id", flat=True)))

    def test_queryset_normaluser(self, user, product_factory, person_factory):
        url = reverse("wbportfolio:product-list")
        request = RequestFactory().get(url)
        permission = Permission.objects.get(codename="view_product", content_type__app_label="wbportfolio")
        user.user_permissions.add(permission)
        force_authenticate(request, user=user)
        view = ProductModelViewSet.as_view({"get": "list"})
        public_product = product_factory.create()
        white_label_product = product_factory.create()
        white_label_product.white_label_customers.add(user.profile)

        other_client_product = product_factory.create()
        other_client_product.white_label_customers.add(person_factory.create())

        response = view(request)
        assert response.status_code == 200
        ids = [r["id"] for r in response.data["results"]]
        assert set(ids) == set([public_product.id, white_label_product.id])



from contextlib import suppress
from datetime import date

from pandas._libs.tslibs.offsets import BDay
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbfdm.models import Instrument
from wbfdm.serializers import InvestableUniverseRepresentationSerializer

from wbportfolio.models import AssetPosition, Portfolio
from wbportfolio.serializers import PortfolioRepresentationSerializer, RebalancerModelSerializer
from wbportfolio.viewsets.configs.display.rebalancing import RebalancerDisplayConfig


class CreateModelPortfolioSerializer(wb_serializers.ModelSerializer):
    create_index = wb_serializers.BooleanField(default=False, label="Create Underlying Index")
    name = wb_serializers.CharField(required=True)
    _currency = CurrencyRepresentationSerializer(source="currency")

    class Meta:
        model = Portfolio
        fields = (
            "name",
            "currency",
            "create_index",
            "_currency",
        )


class AdjustQuoteSerializer(wb_serializers.Serializer):
    old_quote = wb_serializers.PrimaryKeyRelatedField(queryset=Instrument.objects.all())
    _old_quote = InvestableUniverseRepresentationSerializer(source="old_quote")

    new_quote = wb_serializers.PrimaryKeyRelatedField(queryset=Instrument.objects.all())
    _new_quote = InvestableUniverseRepresentationSerializer(
        source="new_quote",
        optional_get_parameters={"old_quote": "sibling_of"},
        depends_on=[{"field": "old_quote", "options": {}}],
    )

    only_portfolios = wb_serializers.PrimaryKeyRelatedField(queryset=Portfolio.objects.all(), required=False)
    _only_portfolios = PortfolioRepresentationSerializer(
        source="only_portfolios",
        optional_get_parameters={"old_quote": "invests_in"},
        depends_on=[{"field": "new_quote", "options": {}}],
        many=True,
    )

    adjust_after = wb_serializers.DateField(required=False)


def _get_portfolio_start_end_serializer_class(portfolio):
    today = date.today()

    class StartEndDateSerializer(wb_serializers.Serializer):
        start = wb_serializers.DateField(
            label="Start",
            default=portfolio.assets.earliest("date").date if portfolio.assets.exists() else today,
        )
        end = wb_serializers.DateField(
            label="End", default=portfolio.assets.latest("date").date if portfolio.assets.exists() else today
        )

    return StartEndDateSerializer


def _get_rebalance_serializer_class(portfolio):
    try:
        default_trade_date = (portfolio.assets.latest("date").date + BDay(1)).date()
    except AssetPosition.DoesNotExist:
        default_trade_date = date.today()

    class RebalanceSerializer(wb_serializers.Serializer):
        trade_date = wb_serializers.DateField(default=default_trade_date)

    return RebalanceSerializer


class PortfolioButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self) -> set:
        return {
            bt.ActionButton(
                method=RequestType.POST,
                identifiers=("wbportfolio:portfolio",),
                endpoint=reverse("wbportfolio:portfolio-adjustquote", args=[], request=self.request),
                label="Adjust quote",
                serializer=AdjustQuoteSerializer,
                action_label="Action triggered",
                title="Adjust Quote",
                instance_display=create_simple_display(
                    [
                        ["old_quote", "new_quote", "adjust_after"],
                        ["only_portfolios", "only_portfolios", "only_portfolios"],
                    ]
                ),
            ),
        }

    def get_custom_instance_buttons(self):
        admin_buttons = [
            bt.ActionButton(
                method=RequestType.POST,
                identifiers=("wbportfolio:portfolio",),
                key="add_automatic_rebalancer",
                label="Attach Rebalancer",
                serializer=RebalancerModelSerializer,
                action_label="Attach Rebalancer",
                title="Attach Rebalancer",
                instance_display=RebalancerDisplayConfig(
                    self.view, self.request, self.instance
                ).get_instance_display(),
            )
        ]
        buttons = [bt.WidgetButton(key="treegraphchart", label="Visualize Tree Graph Chart")]
        with suppress(Portfolio.DoesNotExist, KeyError):
            portfolio = Portfolio.objects.get(id=self.view.kwargs["pk"])
            buttons.append(
                bt.ActionButton(
                    method=RequestType.POST,
                    identifiers=("wbportfolio:portfolio",),
                    key="rebalance",
                    label="Rebalance",
                    serializer=_get_rebalance_serializer_class(portfolio),
                    action_label="Rebalance",
                    title="Rebalance",
                    instance_display=create_simple_display([["trade_date"]]),
                )
            )

            if primary_portfolio := portfolio.primary_portfolio:
                admin_buttons.append(
                    bt.ActionButton(
                        method=RequestType.POST,
                        identifiers=("wbportfolio:portfolio",),
                        key="recompute_lookthrough",
                        label="Recompute Look-Through Portfolio",
                        serializer=_get_portfolio_start_end_serializer_class(primary_portfolio),
                        action_label="Recompute Look-Through Portfolio",
                        title="Recompute Look-Through Portfolio",
                        instance_display=create_simple_display([["start", "end"]]),
                    )
                )
            buttons.append(
                bt.DropDownButton(label="Admin", icon=WBIcon.UNFOLD.icon, buttons=tuple(admin_buttons)),
            )
        return set(buttons)

    def get_custom_list_instance_buttons(self):
        return self.get_custom_instance_buttons()

    # def get_custom_buttons(self):
    #     if not self.view.kwargs.get("pk", None):
    #         return {
    #             bt.ActionButton(
    #                 method=RequestType.POST,
    #                 identifiers=("wbportfolio:portfolio",),
    #                 endpoint=reverse("wbportfolio:portfolio-createmodelportfolio", request=self.request),
    #                 label="Create New Model Portfolio",
    #                 serializer=CreateModelPortfolioSerializer,
    #                 action_label="create",
    #                 title="Create Model Portfolio",
    #                 instance_display=create_simple_display([["name", "currency"], ["create_index", "."]]),
    #             )
    #         }
    #     return set()



import random
from datetime import timedelta
from decimal import Decimal

import pytest
from django.contrib.auth.models import Permission
from faker import Faker
from wbcore.contrib.authentication.models.users import User

from wbportfolio.models import Claim, Trade

fake = Faker()


@pytest.mark.django_db
class TestClaimModel:
    def test_init(self, claim):
        assert claim.id is not None

    def test_filter_claims_for_user(self, user, claim_factory, account_factory, account_role_factory):
        no_account_claims = claim_factory.create(account=None)
        private_claim = claim_factory.create(account=account_factory.create(is_public=False))
        public_claim = claim_factory.create(account=account_role_factory.create(entry=user.profile.entry_ptr).account)
        # if normal user, can only see its own claim, with an attached account
        assert set(Claim.objects.filter_for_user(user)) == {public_claim}
        user.is_internal = True
        user.save()
        # if insternal user, they can also see claim without account
        assert set(Claim.objects.filter_for_user(user)) == {no_account_claims, public_claim}
        user.user_permissions.add(
            Permission.objects.get(content_type__app_label="wbcrm", codename="administrate_account")
        )
        user = User.objects.get(id=user.id)  # we refetch user to clear perm cache, doesn't happen with refresh_from_db
        # administrator can see everything
        assert set(Claim.objects.filter_for_user(user)) == {no_account_claims, public_claim, private_claim}

    def test_account_merging(self, account_factory, claim):
        base_account = account_factory.create()
        merged_account = claim.account
        assert base_account.reference_id != merged_account.reference_id
        base_account.merge(merged_account)
        claim.refresh_from_db()
        assert claim.account == base_account
        assert claim.reference == str(merged_account.reference_id)

    @pytest.mark.parametrize(
        "shares,unvalid_status",
        [
            (
                Decimal(50),
                random.choice(
                    [Claim.Status.WITHDRAWN, Claim.Status.AUTO_MATCHED, Claim.Status.PENDING, Claim.Status.APPROVED]
                ),
            )
        ],
    )
    def test_auto_match_on_claim_save(
        self, weekday, shares, unvalid_status, product, claim_factory, customer_trade_factory
    ):
        valid_trade = customer_trade_factory.create(
            underlying_instrument=product,
            portfolio=product.primary_portfolio,
            transaction_date=fake.date_between(
                weekday - timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                weekday + timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
            ),
            shares=fake.pydecimal(min_value=int(shares - 1), max_value=int(shares + 1)),
        )
        claim = claim_factory.create(product=product, date=weekday, shares=shares, trade=None, status=unvalid_status)
        assert claim.status == unvalid_status
        assert claim.trade is None

        claim = claim_factory.create(
            product=product, date=weekday, shares=shares, trade=None, status=Claim.Status.DRAFT
        )
        assert claim.status == Claim.Status.AUTO_MATCHED
        assert claim.trade == valid_trade

    @pytest.mark.parametrize(
        "shares,unvalid_status",
        [
            (
                Decimal(50),
                random.choice(
                    [Claim.Status.WITHDRAWN, Claim.Status.AUTO_MATCHED, Claim.Status.PENDING, Claim.Status.APPROVED]
                ),
            )
        ],
    )
    def test_claim_auto_match_on_trade_creation(
        self, weekday, shares, unvalid_status, product, claim_factory, customer_trade_factory
    ):
        claim = claim_factory.create(product=product, date=weekday, shares=shares, trade=None, status=unvalid_status)
        trade = customer_trade_factory.create(
            underlying_instrument=product,
            portfolio=product.primary_portfolio,
            transaction_date=fake.date_between(
                weekday - timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                weekday + timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
            ),
            shares=fake.pydecimal(min_value=int(shares - 1), max_value=int(shares + 1)),
        )
        assert claim.status == unvalid_status
        assert claim.trade is None

        # we delete the trade and set the claim status to draft, where we expect a newly created trade to be automatch
        trade.delete()
        claim.status = Claim.Status.DRAFT
        claim.save()
        trade = customer_trade_factory.create(
            underlying_instrument=product,
            portfolio=product.primary_portfolio,
            transaction_date=fake.date_between(
                weekday - timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                weekday + timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
            ),
            shares=fake.pydecimal(min_value=int(shares - 1), max_value=int(shares + 1)),
        )
        claim.refresh_from_db()
        assert claim.status == Claim.Status.AUTO_MATCHED
        assert claim.trade == trade



from wbcore import filters as wb_filters

from wbportfolio.models import Order


class OrderFilterSet(wb_filters.FilterSet):
    has_warnings = wb_filters.BooleanFilter()

    class Meta:
        model = Order
        fields = {"underlying_instrument": ["exact"], "order_type": ["exact"]}



from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbfdm.models import Instrument


class TradeButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {bt.WidgetButton(key="claims", label="Claims", icon=WBIcon.TRADE.icon)}

    def get_custom_instance_buttons(self):
        btn = self.get_custom_list_instance_buttons()
        btn.add(
            bt.HyperlinkButton(
                key="import_source",
                label="Import Source",
                icon=WBIcon.SAVE.icon,
            )
        )
        return btn


class TradeInstrumentButtonConfig(TradeButtonConfig):
    def get_custom_buttons(self):
        res = set()
        if instrument_id := self.view.kwargs.get("instrument_id", None):
            instrument = Instrument.objects.get(id=instrument_id)
            res = {
                bt.WidgetButton(
                    endpoint=reverse(
                        "wbportfolio:instrument-custodiandistribution-list", args=[instrument_id], request=self.request
                    ),
                    label="Custodian Distribution",
                ),
                bt.WidgetButton(
                    endpoint=reverse(
                        "wbportfolio:instrument-customerdistribution-list", args=[instrument_id], request=self.request
                    ),
                    label="Customer Distribution",
                ),
            }
            if instrument.security_instrument_type.key == "product":
                res.add(
                    bt.WidgetButton(
                        endpoint=reverse(
                            "wbportfolio:product-nominalchart-list", args=[instrument_id], request=self.request
                        ),
                        label="Nominal Chart",
                    )
                )
                res.add(
                    bt.WidgetButton(
                        endpoint=reverse(
                            "wbportfolio:product-aumchart-list", args=[instrument_id], request=self.request
                        ),
                        label="AUM Chart",
                    )
                )
        return res



from datetime import date, datetime, timedelta
from json import JSONDecodeError

import requests
from django.conf import settings
from django.utils import timezone
from requests import HTTPError


class UBSNeoAPIClient:
    BASE_URL = "https://neo.ubs.com/api"
    SUCCESS_VALUE = "SUCCESS"
    VIRTUAL_AMC_ISIN = "TEST_API_001"

    def __init__(
        self,
        initial_jwt_token: str,
        jwt_token_expiry_timestamp: datetime | None = None,
    ):
        """
        :param service_account_id: Identifier for your UBS service account (for reference).
        :param initial_jwt_token: JWT token string initially provided to authenticate API calls.
        :param jwt_token_expiry_timestamp: UNIX timestamp when the token expires - to manage renewal.
        """
        self.jwt_token = initial_jwt_token
        self.jwt_token_expiry = (
            jwt_token_expiry_timestamp if jwt_token_expiry_timestamp else timezone.now() + timedelta(days=1)
        )

    def _is_token_expired(self) -> bool:
        """Check if the current token is expired or near expiry (e.g. within 5 minutes)."""
        return timezone.now() > self.jwt_token_expiry - timedelta(minutes=5)

    def _renew_token(self):
        """
        Placeholder: Implement token renewal logic here.
        This usually involves interacting with UBS Neo application or token service
        before expiry to obtain a new token. This must be customized per your process.
        """
        raise ValueError("Token has expired. Please go to https://neo.ubs.com/ and renew it.")

    def _get_headers(self) -> dict[str, str]:
        """Prepare HTTP headers including Authorization bearer token."""
        if self._is_token_expired():
            self._renew_token()
        return {"Authorization": f"Bearer {self.jwt_token}", "Content-Type": "application/json"}

    def _get_json(self, response) -> dict:
        try:
            return response.json()
        except JSONDecodeError:
            return dict()

    def _validate_isin(self, isin: str, test: bool = False) -> str:
        # ensure the given isin can be used for rebalancing (e.g. debug or dev mode)
        if test or settings.DEBUG:
            return self.VIRTUAL_AMC_ISIN
        return isin

    def _raise_for_status(self, response):
        try:
            response.raise_for_status()
        except HTTPError as e:
            json_response = response.json()
            raise HTTPError(json_response.get("errors", json_response.get("message"))) from e

    def get_rebalance_service_status(self) -> dict:
        """Check API connection status."""
        url = f"{self.BASE_URL}/ged-amc/external/rebalance/v1/status"
        response = requests.get(url, headers=self._get_headers(), timeout=10)
        self._raise_for_status(response)
        return self._get_json(response)

    def get_rebalance_status_for_isin(self, isin: str) -> dict:
        """Check certificate accessibility and workflow status for given ISIN."""
        url = f"{self.BASE_URL}/ged-amc/external/rebalance/v1/status/{isin}"
        response = requests.get(url, headers=self._get_headers(), timeout=10)
        self._raise_for_status(response)
        return self._get_json(response)

    def submit_rebalance(self, isin: str, items: list[dict[str, str]], test: bool = False) -> dict:
        """
        Submit a rebalance request.
        :param isin: Certificate ISIN string.
        :param orders: List of dto representing order instructions.
        :param test: If True, submits to test endpoint to validate syntax only (no persistence).
        """
        isin = self._validate_isin(isin, test=test)
        url = f"{self.BASE_URL}/ged-amc/external/rebalance/v1/submit/{isin}"
        payload = {"items": items}
        response = requests.post(url, json=payload, headers=self._get_headers(), timeout=60)
        self._raise_for_status(response)
        return self._get_json(response)

    def save_draft(self, isin: str, items: list[dict[str, str]]) -> dict:
        """Save a rebalance draft."""
        isin = self._validate_isin(isin)
        url = f"{self.BASE_URL}/ged-amc/external/rebalance/v1/savedraft/{isin}"
        payload = {"items": items}
        response = requests.post(url, json=payload, headers=self._get_headers(), timeout=60)
        self._raise_for_status(response)
        return self._get_json(response)

    def cancel_rebalance(self, isin: str) -> dict:
        """Cancel a rebalance request."""
        isin = self._validate_isin(isin)
        url = f"{self.BASE_URL}/ged-amc/external/rebalance/v1/cancel/{isin}"
        response = requests.delete(url, headers=self._get_headers(), timeout=60)
        self._raise_for_status(response)
        return self._get_json(response)

    def get_current_rebalance_request(self, isin: str) -> dict:
        """Fetch the current rebalance request for a certificate."""
        url = f"{self.BASE_URL}/ged-amc/external/rebalance/v1/currentRebalanceRequest/{isin}"
        response = requests.get(url, headers=self._get_headers(), timeout=60)
        self._raise_for_status(response)
        return self._get_json(response)

    def get_rebalance_reports(self, isin: str, from_date: date, to_date: date) -> dict:
        """Fetch the current rebalance request for a certificate."""
        url = f"{self.BASE_URL}/ged-amc/external/report/v1/rebalance/{isin}"
        response = requests.get(
            url,
            headers=self._get_headers(),
            timeout=60,
            params={"fromDate": from_date.strftime("%Y-%m-%d"), "toDate": to_date.strftime("%Y-%m-%d")},
        )
        self._raise_for_status(response)
        return self._get_json(response)

    def get_portfolio_at_date(self, isin: str, val_date: date) -> dict:
        url = f"https://neo.ubs.com/api/ged-amc/external/report/v1/valuation/{isin}/{val_date:%Y-%m-%d}"
        response = requests.get(url, headers=self._get_headers(), timeout=10)
        self._raise_for_status(response)
        return self._get_json(response)

    def get_management_fees(self, isin: str, from_date: date, to_date: date) -> dict:
        url = f"https://neo.ubs.com/api/ged-amc/external/fee/v1/management/{isin}"
        response = requests.get(
            url,
            headers=self._get_headers(),
            params={"fromDate": from_date.strftime("%Y-%m-%d"), "toDate": to_date.strftime("%Y-%m-%d")},
            timeout=30,
        )
        self._raise_for_status(response)
        return self._get_json(response)

    def get_performance_fees(self, isin: str, from_date: date, to_date: date) -> dict:
        url = f"https://neo.ubs.com/api/ged-amc/external/fee/v1/performance/{isin}"
        response = requests.get(
            url,
            headers=self._get_headers(),
            params={"fromDate": from_date.strftime("%Y-%m-%d"), "toDate": to_date.strftime("%Y-%m-%d")},
            timeout=30,
        )
        self._raise_for_status(response)
        return self._get_json(response)

    def validate_response(self, response: dict) -> dict:
        if response.get("status", "") == self.SUCCESS_VALUE:
            return response
        return dict()



from datetime import timedelta
from decimal import Decimal

import pytest
from wbfdm.models import InstrumentPrice

from wbportfolio.models import FeeCalculation, Fees, Product


def fees_calculation(price_id):
    price = InstrumentPrice.objects.get(id=price_id)
    product = Product.objects.get(id=price.instrument.id)
    yield {
        "product": product,
        "fee_date": price.date,
        "transaction_subtype": Fees.Type.MANAGEMENT,
        "currency": product.currency,
        "calculated": True,
        "total_value": price.net_value,
        "total_value_gross": price.net_value,
    }


@pytest.mark.django_db
class TestFeesModel:
    def test_init(self, fees):
        assert fees.id is not None

    def test_filter_only_valid_fees(self, fees_factory):
        fees_d0 = (
            fees_factory.create()
        )  # no matter if estimated or not, we expect this fee to be on the resulting queryset
        calculated_fees_d1 = fees_factory.create(  # there will be a real fee for that date, type and product, so this will be filtered out
            calculated=True,
            product=fees_d0.product,
            fee_date=(fees_d0.fee_date + timedelta(days=1)),
            transaction_subtype=fees_d0.transaction_subtype,
        )
        real_fees_d1 = fees_factory.create(
            calculated=False,
            product=calculated_fees_d1.product,
            fee_date=calculated_fees_d1.fee_date,
            transaction_subtype=calculated_fees_d1.transaction_subtype,
        )

        assert set(Fees.valid_objects.all()) == {fees_d0, real_fees_d1}
        assert set(Fees.objects.filter_only_valid_fees()) == {fees_d0, real_fees_d1}

    def test_compute_fee_from_price(self, product_factory, instrument_price_factory):
        # we check the mechanism of computing fees on valuation creation
        fee_calculation = FeeCalculation.objects.create(
            name="base computation", import_path="wbportfolio.tests.models.transactions.test_fees"
        )
        product = product_factory.create(fee_calculation=fee_calculation)
        price = instrument_price_factory.create(instrument=product)  # post save must be called to compute fees

        fees = Fees.objects.get(
            product=product, fee_date=price.date, calculated=True, transaction_subtype=Fees.Type.MANAGEMENT
        )
        assert fees.total_value == pytest.approx(price.net_value, rel=Decimal(1e-4))



from datetime import date, timedelta
from datetime import date as date_lib
from decimal import Decimal

from django.db import models
from django.db.models import (
    DateField,
    Exists,
    ExpressionWrapper,
    OuterRef,
    Q,
    QuerySet,
    Sum,
)
from django.db.models.functions import Greatest
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from wbcore.contrib.ai.llm.config import add_llm_prompt
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.models import WBModel
from wbcore.signals import pre_merge
from wbcore.signals.models import pre_collection
from wbcore.utils.enum import ChoiceEnum
from wbcore.utils.strings import ReferenceIDMixin
from wbcrm.models import AccountRole
from wbcrm.models.accounts import Account
from wbfdm.models.instruments.instrument_prices import InstrumentPrice

from ..custodians import Custodian
from ..llm.wbcrm.analyze_relationship import get_performances_prompt
from ..roles import PortfolioRole
from .trades import Trade


def can_administrate_claim(claim, user):
    today = date_lib.today()
    return (
        user.profile.portfolio_roles.filter(
            Q(role_type=PortfolioRole.RoleType.MANAGER)
            & (Q(start__isnull=True) | Q(start__lte=today))
            & (Q(end__isnull=True) | Q(end__gte=today))
        ).exists()
        or user.is_superuser
    )


def can_edit_claim(claim, user):
    return (claim.claimant and claim.claimant.id == user.profile.id) or user.is_internal or user.is_superuser


class ClaimGroupbyChoice(ChoiceEnum):
    ROOT_ACCOUNT = "Root Account"
    ACCOUNT = "Account"
    ROOT_ACCOUNT_OWNER = "Root Account Owner"
    ACCOUNT_OWNER = "Account Owner"
    PRODUCT = "Product"
    PRODUCT_GROUP = "ProductGroup"
    CLASSIFICATION = "Classification"

    @classmethod
    @property
    def map(cls):
        return {
            "ROOT_ACCOUNT": {
                "pk": "root_account",
                "title_key": "root_account_repr",
                "search_fields": ["root_account_repr"],
            },
            "ACCOUNT": {
                "pk": "account",
                "title_key": "account__computed_str",
                "search_fields": ["account__computed_str"],
            },
            "ROOT_ACCOUNT_OWNER": {
                "pk": "root_account_owner",
                "title_key": "root_account_owner_repr",
                "search_fields": ["root_account_owner_repr"],
            },
            "ACCOUNT_OWNER": {
                "pk": "account__owner",
                "title_key": "account__owner__computed_str",
                "search_fields": ["account__owner__computed_str"],
            },
            "PRODUCT": {
                "pk": "product",
                "title_key": "product__computed_str",
                "search_fields": ["product__computed_str"],
            },
            "PRODUCT_GROUP": {
                "pk": "product__parent",
                "title_key": "product__parent__name",
                "search_fields": ["product__parent__name"],
            },
            "CLASSIFICATION": {
                "pk": "classification_id",
                "title_key": "classification_title",
                "search_fields": ["classification_title"],
            },
        }

    @classmethod
    def get_map(cls, name):
        return cls.map[name]


class ClaimDefaultQueryset(QuerySet):
    def filter_for_user(self, user: User, validity_date: date_lib | None = None) -> QuerySet:
        """
        Protect the chained queryset and filter the claims that this user cannot see based on the following rules:
        """
        if user.has_perm("wbcrm.administrate_account"):
            return self
        allowed_accounts = Account.objects.filter_for_user(user, validity_date=validity_date).values(
            "id"
        )  # This is way faster
        # accounts = self.annotate(can_see_account=Exists(allowed_accounts.filter(id=OuterRef("account"))))
        if user.is_internal:
            return self.filter(Q(account__isnull=True) | Q(account__in=allowed_accounts))
        return self.filter(Q(account__in=allowed_accounts) | (Q(account__isnull=True) & Q(creator_id=user.profile.id)))

    def filter_for_customer(self, customer: Entry, include_related_roles: bool = False) -> QuerySet:
        """
        Filter the chained queryset to return only the claim that belongs to a certain customer
        """
        customer_account_ids = Account.get_accounts_for_customer(customer).values("id")

        if not include_related_roles:
            return self.filter(account__in=customer_account_ids)
        return self.annotate(
            role_exists=Exists(AccountRole.objects.filter(entry=customer, account=OuterRef("account")))
        ).filter(Q(account__in=customer_account_ids) | Q(role_exists=True))

    def annotate_asset_under_management_for_date(self, val_date: date):
        return self.annotate(
            net_value=InstrumentPrice.subquery_closest_value("net_value", val_date, instrument_pk_name="product"),
            fx_rate=CurrencyFXRates.get_fx_rates_subquery(val_date),
            asset_under_management=models.ExpressionWrapper(
                models.F("net_value") * models.F("shares"),
                output_field=models.DecimalField(max_digits=16, decimal_places=4, default=0.0),
            ),
            asset_under_management_usd=models.ExpressionWrapper(
                models.F("asset_under_management") * models.F("fx_rate"),
                output_field=models.DecimalField(max_digits=16, decimal_places=4, default=0.0),
            ),
        )


class ClaimManager(models.Manager):
    def get_queryset(self) -> ClaimDefaultQueryset:
        return ClaimDefaultQueryset(self.model)

    def filter_for_user(self, user: User, validity_date: date_lib | None = None) -> QuerySet:
        return self.get_queryset().filter_for_user(user, validity_date=validity_date)

    def filter_for_customer(self, customer: Entry, include_related_roles: bool = False) -> QuerySet:
        return self.get_queryset().filter_for_customer(customer)

    def annotate_asset_under_management_for_date(self, val_date: date) -> QuerySet:
        return self.get_queryset().annotate_asset_under_management_for_date(val_date)


# @workflow.register(serializer_class="wbportfolio.serializers.transactions.claim.ClaimModelSerializer") #we don't register for now. Uncomment as soon as we want to enable workflow on that model
class Claim(ReferenceIDMixin, WBModel):
    """A customer can claim that a trade or part of a trade was executed my them"""

    class Status(models.TextChoices):
        PENDING = "PENDING", "Pending"
        APPROVED = "APPROVED", "Approved"
        WITHDRAWN = "WITHDRAWN", "Withdrawn"
        DRAFT = "DRAFT", "Draft"
        AUTO_MATCHED = "AUTO_MATCHED", "Auto-Matched"

    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name="Status")
    account = models.ForeignKey(
        to="wbcrm.Account",
        related_name="claims",
        null=True,
        blank=True,
        limit_choices_to=models.Q(is_terminal_account=True),
        on_delete=models.SET_NULL,
        verbose_name="Account",
        # help_text="The account the claim is assigned to. If no sub-account is provided it will be tried to be assigned to a sub account based on the given claimant."
    )

    trade = models.ForeignKey(
        to="wbportfolio.Trade",
        related_name="claims",
        blank=True,
        null=True,
        on_delete=models.PROTECT,  # We protect the claim in case of trade deletion. This needs to be handled upstream
        verbose_name="Trade",
        help_text="Please select a Product first. The customer-trade the claim is consolidated against. The customer-trade and the claim don't necessarily have to have the same date, number of shares, etc.",
    )
    product = models.ForeignKey(
        to="wbportfolio.Product", related_name="claims", on_delete=models.PROTECT, verbose_name="Product"
    )
    claimant = models.ForeignKey(
        to="directory.Entry",
        related_name="claimed",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Claimant",
        # help_text="The person / company that claims this trade. If no claimant is given, it will assigned to the current user."
    )
    creator = models.ForeignKey(
        to="directory.Entry",
        related_name="claims_created",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Creator",
    )
    date = models.DateField(verbose_name="Trade Date")
    bank = models.CharField(max_length=255, blank=True, verbose_name="Bank")
    reference = models.CharField(max_length=255, null=True, blank=True, verbose_name="Additional Reference")
    shares = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        help_text="The amount of shares purchased / sold.",
        null=True,
        blank=True,
        verbose_name="Shares",
    )

    nominal_amount = models.DecimalField(
        max_digits=15,
        decimal_places=2,
        help_text="The nominal amount purchased / sold. Either shares or nominal_amount has to be provided.",
        null=True,
        blank=True,
        verbose_name="Nominal Amount",
    )
    external_id = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="An External ID to reference a claim.",
        verbose_name="External Identifier",
    )
    as_shares = models.BooleanField(default=True)

    objects = ClaimManager()

    class Meta:
        verbose_name = "Claim"
        verbose_name_plural = "Claims"

    def __str__(self):
        return f"{self.reference_id} {self.product.name} ({self.bank} - {self.shares:,} shares - {self.date}) "

    def save(self, *args, auto_match: bool = True, **kwargs):
        if self.shares is None and self.nominal_amount is None:
            raise ValueError(
                f"Either shares or nominal amount have to be provided. Shares={self.shares}, Nominal={self.nominal_amount}"
            )
        if self.product:
            if self.shares is not None:
                self.nominal_amount = self.shares * self.product.share_price
            else:
                self.shares = self.nominal_amount / self.product.share_price
        if not self.trade and self.status == self.Status.DRAFT and auto_match:
            self.trade = self.auto_match()
        # we shouldn't remove the link to the trade even if the claim is withdrawn for audit purposes, this might screw any reverse queryset such as claims.all() if widthdrawn claims are not explicitly excluded.
        # if self.status == self.Status.WITHDRAWN and self.trade:
        #     self.trade = None
        if self.trade:
            if not self.trade.is_claimable:
                raise ValueError("The selected trade is not a valid customer trade")
            if self.product and self.trade.underlying_instrument.id != self.product.id:
                raise ValueError("The selected product does not match the trade underlying instrument")
        super().save(*args, **kwargs)

    @classmethod
    def get_valid_and_approved_claims(cls, val_date: date_lib | None = None, account: Account | None = None):
        claims = cls.objects.annotate(
            date_considered=ExpressionWrapper(
                Greatest("trade__transaction_date", "date") + 1, output_field=DateField()
            )
        ).filter(Q(account__is_terminal_account=True) & Q(account__is_active=True) & Q(status=cls.Status.APPROVED))
        if val_date:
            claims = claims.filter(date_considered__lt=val_date)
        if account:
            claims = claims.filter(account__in=account.get_descendants(include_self=True))
        return claims

    @property
    def nominal_value(self):
        """Returns the nominal value of a claim

        Shares x Share Price

        Returns:
            Decimal -- Nominal Value
        """
        return self.shares * self.product.share_price

    @transition(
        status,
        [Status.PENDING, Status.AUTO_MATCHED],
        Status.APPROVED,
        permission=can_administrate_claim,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:claim",),
                icon=WBIcon.APPROVE.icon,
                color=ButtonDefaultColor.SUCCESS,
                key="approve",
                label="Approve",
                action_label="Approve",
                description_fields="<p>Date: {{date}}</p><p>Shares: {{shares}}</p><p>Product: {{_product.name}} ({{_product.isin}})</p>",
                instance_display=create_simple_display(
                    [
                        ["date", "shares", "bank"],
                        ["product", "claimant", "."],
                        ["account", "account", "account"],
                        ["trade", "trade", "trade"],
                    ]
                ),
            )
        },
    )
    def approve(self, **kwargs):
        pass

    def can_approve(self):
        errors = dict()

        if not self.trade:
            errors["trade"] = [_("With this status, this has to be provided.")]

        if not self.product:
            errors["product"] = [_("With this status, this has to be provided.")]

        if not self.account:
            errors["account"] = [_("With this status, this has to be provided.")]

        # check if the specified product have a valid nav at the specified date
        if (
            (product := self.product)
            and (claim_date := self.date)
            and not product.valuations.filter(date=claim_date).exists()
        ):
            if (prices_qs := product.valuations.filter(date__lt=claim_date)).exists():
                errors["date"] = [
                    f"For product {product.name}, the latest valid valuation date before {claim_date:%Y-%m-%d} is {prices_qs.latest('date').date:%Y-%m-%d}: Please select a valid date."
                ]
            else:
                errors["date"] = [
                    f"There is no valuation before {claim_date:%Y-%m-%d} for product {product.name}: Please select a valid date."
                ]
        return errors

    @transition(
        status,
        [Status.PENDING, Status.AUTO_MATCHED],
        Status.DRAFT,
        permission=lambda claim, user: user.is_internal or user.is_superuser,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.ERROR,
                identifiers=("wbportfolio:claim",),
                icon=WBIcon.UNDO.icon,
                key="backtodraft",
                label="Back to Draft",
                action_label="Back to Draft",
                description_fields="<p>Date: {{date}}</p><p>Shares: {{shares}}</p>",
            )
        },
    )
    def backtodraft(self, **kwargs):
        pass

    @transition(
        status,
        [Status.DRAFT],
        Status.WITHDRAWN,
        permission=can_edit_claim,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:claim",),
                color=ButtonDefaultColor.ERROR,
                icon=WBIcon.DELETE.icon,
                key="withdraw",
                label="Withdraw",
                action_label="Withdraw",
                description_fields="<p>Date: {{date}}</p><p>Shares: {{shares}}</p>",
            )
        },
    )
    def withdraw(self, **kwargs):
        pass

    @transition(
        status,
        [Status.APPROVED],
        Status.DRAFT,
        permission=lambda claim, user: user.is_internal or user.is_superuser,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:claim",),
                icon=WBIcon.EDIT.icon,
                color=ButtonDefaultColor.WARNING,
                key="revise",
                label="Revise",
                action_label="Revise",
                description_fields="<p>Date: {{date}}</p><p>Product: {{_product.name}}</p>",
            )
        },
    )
    def revise(self, **kwargs):
        pass

    @transition(
        status,
        [Status.DRAFT, Status.AUTO_MATCHED],
        Status.PENDING,
        permission=lambda claim, user: user.is_internal or user.is_superuser,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:claim",),
                icon=WBIcon.SEND.icon,
                color=ButtonDefaultColor.WARNING,
                key="submit",
                label="Submit for approval",
                action_label="Submit",
                description_fields="<p>Date: {{date}}</p><p>Product: {{_product.name}}</p>",
            )
        },
    )
    def submit(self, **kwargs):
        pass

    def can_submit(self):
        return self.can_approve()

    def auto_match(self) -> Trade | None:
        shares_epsilon = 1  # share
        auto_match_trade = None
        # Obvious filtering
        trades = Trade.valid_customer_trade_objects.filter(
            Q(transaction_date__gte=self.date - timedelta(days=Trade.TRADE_WINDOW_INTERVAL))
            & Q(transaction_date__lte=self.date + timedelta(days=Trade.TRADE_WINDOW_INTERVAL))
        )
        if self.product:
            trades = trades.filter(underlying_instrument=self.product)
        # Find trades by shares (or remaining to be claimed)
        trades = trades.filter(
            Q(diff_shares__lte=self.shares + shares_epsilon) & Q(diff_shares__gte=self.shares - shares_epsilon)
        )
        if trades.count() == 1:
            auto_match_trade = trades.first()

        # Find trades by custodian
        if (
            not auto_match_trade
            and self.bank
            and (custodian := Custodian.get_by_mapping(self.bank, use_similarity=True, create_missing=False))
        ):
            if trades.filter(custodian=custodian).exists():
                trades = trades.filter(custodian=custodian)
            if trades.count() == 1:
                auto_match_trade = trades.first()

        # Find trades by external_id
        if not auto_match_trade and self.external_id and trades.count() > 1:
            trades = trades.filter(
                Q(external_id__icontains=self.external_id) | Q(external_id_alternative__icontains=self.external_id)
            )
            if trades.count() == 1:
                auto_match_trade = trades.first()

        if auto_match_trade:
            self.status = self.Status.AUTO_MATCHED
            return auto_match_trade

    @classmethod
    def subquery_assets_under_management_for_account(cls, claims, price_date, account_key="account__pk"):
        """Returns a subquery which annotates the assets in USD for each Sub Account based on a given queryset

        Arguments:
            claims {QuerySet<commission.Claim>} -- Prefiltered queryset of claims
            price_date {datetime.date} -- the date which is used to calculate everything
            account_key {str} -- the outer reference to the sub account pk (default: account__pk)

        Returns:
            Subquery -- Sub Account with assets in USD
        """

        return models.Subquery(
            claims.filter(
                status=cls.Status.APPROVED,
                account__pk=models.OuterRef(account_key),
                trade__transaction_date__lte=price_date,
            )
            .annotate_asset_under_management_for_date(price_date)
            .values("account")
            .annotate(sum_assets_usd=models.Sum("assets_under_management_usd"))
            .values("sum_assets_usd")[:1],
            output_field=models.DecimalField(max_digits=16, decimal_places=4, default=0.0),
        )

    @classmethod
    def subquery_claim_sum_for_account(cls, claims, price_date, account_key="account__pk"):
        return models.Subquery(
            claims.filter(
                status=cls.Status.APPROVED,
                account__pk=models.OuterRef(account_key),
                trade__transaction_date__lte=price_date,
            )
            .values("account")
            .annotate(claim_sum=models.Sum("shares"))
            .values("claim_sum")[:1],
            output_field=models.DecimalField(max_digits=16, decimal_places=0, default=0.0),
        )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:claim"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:claimrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{shares}} ({{account}})"


@receiver(models.signals.post_save, sender="wbportfolio.Trade")
def post_save_trade(sender, instance, created, raw, **kwargs):
    """
    For every claimable trade, first try to autoclaim the already existing draft claims with this new trade. And auto claim them as well if the product has a default sub account
    """
    if not raw and instance.is_claimable and created:
        min_transaction_date = instance.transaction_date - timedelta(days=Trade.TRADE_WINDOW_INTERVAL)
        max_transaction_date = instance.transaction_date + timedelta(days=Trade.TRADE_WINDOW_INTERVAL)
        unlinked_claims = Claim.objects.filter(
            status=Claim.Status.DRAFT,
            date__gte=min_transaction_date,
            date__lte=max_transaction_date,
            trade__isnull=True,
            product__id=instance.underlying_instrument.id,
        )
        for claim in unlinked_claims.all():
            claim.trade = claim.auto_match()
            claim.save()
        if instance.product and (account := instance.product.default_account):
            shares = instance.shares or Decimal(0)
            claimed_shares = instance.claims.filter(status=Claim.Status.APPROVED).aggregate(s=models.Sum("shares"))[
                "s"
            ] or Decimal(0)
            rest_shares = shares - claimed_shares
            if rest_shares != 0:
                Claim.objects.create(
                    trade=instance,
                    shares=rest_shares,
                    bank=instance.bank,
                    date=instance.transaction_date,
                    product=instance.product,
                    account=account,
                    claimant=account.owner,
                )
            for claim in instance.claims.exclude(Q(status=Claim.Status.APPROVED) | Q(status=Claim.Status.WITHDRAWN)):
                if claim.status == Claim.Status.DRAFT:
                    claim.submit()
                if claim.status == Claim.Status.PENDING:
                    claim.approve()
                claim.save()


@receiver(pre_collection, sender="wbportfolio.Trade")
def pre_trade_deletion(sender, instance, **kwargs):
    # We can't use pre_delete signal here because the collector is collected before the signal and thus, it will
    for claim in instance.claims.filter(
        status__in=[Claim.Status.DRAFT, Claim.Status.AUTO_MATCHED, Claim.Status.WITHDRAWN]
    ).all():
        claim.trade = None
        claim.save(auto_match=False)

    if (instance.marked_for_deletion or instance.pending) and instance.claims.exists():
        # If a default account exists on the trade's product, it means that the product's trade are automatically claims. In that case, we are sure that a valid trade still exists and the marked for deletion's claim can be safely remove
        if instance.product and instance.product.default_account:
            instance.claims.update(status=Claim.Status.WITHDRAWN)
        else:
            similar_trades = Trade.objects.filter(
                underlying_instrument=instance.underlying_instrument,
                portfolio=instance.portfolio,
                marked_for_deletion=True,
                transaction_date=instance.transaction_date,
            )
            # We check if the sum of marked for deletion trades share sums to 0, in that case, we delete them without regards and set their potential claims to draft
            if similar_trades.exists() and similar_trades.aggregate(s=Sum("shares"))["s"] == 0:
                for t in similar_trades.all():
                    t.claims.exclude(status=Claim.Status.WITHDRAWN).update(trade=None, status=Claim.Status.DRAFT)
            elif (other_unclaims_similar_trades := instance.get_alternative_valid_trades()).exists():
                if other_unclaims_similar_trades.count() > 1:
                    other_unclaims_similar_trades = other_unclaims_similar_trades.filter(bank=instance.bank)
                if other_trade := other_unclaims_similar_trades.first():
                    # If we find an unclaim trade with similar attributes, we forward the marked_for_deletion attribute to it, which will be handled/deleted in a next delete iteration
                    instance.claims.exclude(status=Claim.Status.WITHDRAWN).update(trade=other_trade)


@receiver(pre_merge, sender="wbcrm.Account")
def handle_pre_merge_account_for_claim(
    sender: models.Model, merged_object: "Account", main_object: "Account", **kwargs
):
    """
    Simply reassign the claim linked to the merged account to the main account
    """
    Claim.objects.filter(account=merged_object).update(account=main_object, reference=merged_object.reference_id)


@receiver(add_llm_prompt, sender="wbcrm.Account")
def add_holdings_to_account_heat(sender, instance, key, **kwargs):
    if key == "analyze_relationship":
        return get_performances_prompt(instance)
    return []


@receiver(pre_merge, sender="directory.Entry")
def pre_merge_entry_claim(sender: models.Model, merged_object, main_object, **kwargs):
    Claim.objects.filter(claimant=merged_object).update(claimant=main_object)
    Claim.objects.filter(creator=merged_object).update(creator=main_object)



from django.dispatch import receiver
from wbcore.contrib.directory.viewsets import (
    CompanyModelViewSet,
    EntryModelViewSet,
    PersonModelViewSet,
)
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.display import create_simple_display
from wbcore.signals.instance_buttons import add_instance_button
from wbcrm.viewsets import AccountModelViewSet
from wbfdm.viewsets.instruments import InstrumentModelViewSet

from wbportfolio.serializers.reconciliations import AccountReconciliationModelSerializer

from .mixins import InstrumentButtonMixin


@receiver(add_instance_button, sender=InstrumentModelViewSet)
def add_instrument_request_button(sender, many, *args, request=None, view=None, pk=None, **kwargs):
    return InstrumentButtonMixin.add_instrument_request_button(request=request, view=view, pk=pk)


@receiver(add_instance_button, sender=InstrumentModelViewSet)
def add_transactions_request_button(sender, many, *args, request=None, view=None, pk=None, **kwargs):
    return InstrumentButtonMixin.add_transactions_request_button(request=request, view=view, pk=pk)


@receiver(add_instance_button, sender=PersonModelViewSet)
@receiver(add_instance_button, sender=EntryModelViewSet)
@receiver(add_instance_button, sender=CompanyModelViewSet)
def crm_claim_buttons(sender, many, *args, **kwargs):
    if many:
        return
    return bt.DropDownButton(
        label="Claims",
        icon=WBIcon.UNFOLD.icon,
        buttons=(
            bt.WidgetButton(key="claims", label="Claims", icon=WBIcon.TRADE.icon),
            bt.WidgetButton(key="aum", label="AUM", icon=WBIcon.DOLLAR.icon),
        ),
    )


@receiver(add_instance_button, sender=AccountModelViewSet)
def add_account_holdings_reconciliation_button(sender, many, *args, **kwargs):
    if many:
        return
    return bt.DropDownButton(
        label="Reconciliations",
        icon=WBIcon.UNFOLD.icon,
        buttons=[
            bt.WidgetButton(
                label="Reconciliations",
                key="wbportfolio-accountreconciliation",
            ),
            bt.ActionButton(
                label="Reconcile",
                icon=WBIcon.APPROVE.icon,
                key="wbportfolio-reconcile-account",
                method=RequestType.POST,
                identifiers=("wbportfolio:accountreconciliation", "wbportfolio:accountreconciliationline"),
                action_label="Reconcile this account. A notification will be send to the customer",
                serializer=AccountReconciliationModelSerializer,
                instance_display=create_simple_display([["reconciliation_date"]]),
            ),
        ],
    )



from wbcore.metadata.configs.endpoints import EndpointViewConfig


class AssetPositionPandasEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None


class AggregatedAssetPositionLiquidityEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None



from typing import Optional

from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

perf_formatting_rules = [
    dp.FormattingRule(
        style={
            "color": WBColor.YELLOW_DARK.value,
        },
        condition=(">", 3),
    ),
    dp.FormattingRule(
        style={
            "color": WBColor.RED_DARK.value,
        },
        condition=(">", 5),
    ),
]


class AssetPositionPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label="Title"),
                # dp.Field(key="price_start", label="Price Start"),
                # dp.Field(key="price_end", label="Price End"),
                dp.Field(key="total_value_start", label="Value Start ($)"),
                dp.Field(key="total_value_end", label="Value End ($)"),
                dp.Field(key="allocation_start", label="Weighting Start"),
                dp.Field(key="allocation_end", label="Weighting End"),
                dp.Field(key="performance_total", label="Total Performance"),
                dp.Field(key="performance_forex", label="Forex Performance"),
                dp.Field(key="contribution_total", label="Total Contribution"),
                dp.Field(key="contribution_forex", label="Forex Contribution"),
                dp.Field(key="market_share", label="Market Shares (end)"),
            ]
        )


class AggregatedAssetPositionLiquidityDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        historic_date = self.request.GET.get("historic_date", None)
        compared_date = self.request.GET.get("compared_date", None)
        return dp.ListDisplay(
            fields=[
                dp.Field(
                    key="name",
                    label="Name",
                    formatting_rules=[dp.FormattingRule(style={"fontWeight": "bold"})],
                ),
                dp.Field(
                    key="shares_first_date", label=f"Total Shares - {historic_date}" if historic_date else "None"
                ),
                dp.Field(
                    key="volume_50d_first_date", label=f"Mean Volume - {historic_date}" if historic_date else "None"
                ),
                dp.Field(
                    key="liquidity_first_date",
                    label=f"Days To Liquidate - {historic_date}" if historic_date else "None",
                ),
                dp.Field(
                    key="pct_aum_first_date", label=f"Percent AUM - {historic_date}" if historic_date else "None"
                ),
                dp.Field(
                    key="liquidity_second_date",
                    label=f"Days To Liquidate - {compared_date}" if compared_date else "None",
                ),
                dp.Field(
                    key="pct_aum_second_date", label=f"Percent AUM - {compared_date}" if compared_date else "None"
                ),
            ],
        )



# Author: Skip Montanaro (skip@mojam.com)
# Version: 0.1
# https://github.com/smontanaro/python-bits/blob/master/sylk.py

"""Read SYLK files

Caveat emptor!  This module has only been tested with SYLK files generated
by AppleWorks 5.0!  It almost certainly needs work to be able to process
files generated by other spreadsheets.
"""

import re
from ast import literal_eval


class Table:
    """
    Table of rows used to store sylk parsed datas

    rows are compound of a various amount of datas (different length)
    """

    def __init__(self):
        self.rows = []

    def __setitem__(self, xy_tuple, val):
        """
        Setitem method to set datas in this grid

        :param tuple xy_tuple: The x,y coordinates as 2-uple
        :param str val: The value to store in the grid

        Handle the completion of a row to ensure the grid's row is wide enough
        to set the value
        """
        (x, y) = xy_tuple
        ox = x - 1
        oy = y - 1
        # Ensure the table is long enough to store the val
        while len(self.rows) < y:
            new_row = [" "] * x
            self.rows.append(new_row)

        row = self.rows[oy]
        if val not in ("", " "):
            # Ensure the row is large enough to store the val
            if len(row) < x:
                self._extend_row(row, x - len(row))
            self.rows[oy][ox] = val

    def _extend_row(self, row, missing):
        """
        Extend a row adding void values

        :param list row: a list of datas
        :param int missing: The number of void items to add
        """
        row.extend([" "] * missing)
        return row

    def __iter__(self):
        for row in self.rows:
            yield row


class SYLK:
    """class to read SYLK files and dump to CSV"""

    # when time began
    # different computers use different core dates and store dates as offsets
    # this makes SYLK inherently unportable, but we fudge that by
    # using the ID field to guess at the creating platform
    # note that PCs apparently can't properly decode SYLK files generated
    # on Macs using Appleworks/Clarisworks because they don't take this into
    # account
    unixepoch = (1970, 1, 1, 0, 0, 0, 0, 0, 0)
    macepoch = (1904, 1, 1, 0, 0, 0, 0, 0, 0)
    # this is pure fiction...
    pcepoch = (1900, 1, 1, 0, 0, 0, 0, 0, 0)

    # map SYLK format strings into data types
    knownformats = {
        "General": "string",
        "0": "int",
        "0.00": "float",
        "#,##0": "int",
        "#,##0.00": "float",
        r'"$"#,##0\ ;;\("$"#,##0\,': "float",
        r'"$"#,##0.00\ ;;\("$"#,##0.00\,': "float",
        "0%": "float",
        "0.00%": "float",
        "0.00E+00": "float",
        "m/d/yy": "date",
        "d-mmm-yy": "date",
        "d-mmm": "date",
        "mmm-yy": "date",
        "h:mm AM/PM": "time",
        "h:mm:ss AM/PM": "time",
        "h:mm": "time",
        "h:mm:ss": "time",
        "hh:mm AM/PM": "time",
        "hh:mm:ss AM/PM": "time",
        "m/d/yy h:mm": "datetime",
        "m-dd-yy": "date",
        "m-dd": "date",
        '"$"#,##0 ;;[Red]("$"#,##0,': "float",
        '"$"#,##0.00 ;;[Red]("$"#,##0.00,': "float",
        "mmm d, yyyy": "date",
        "mmmm d, yyyy": "date",
        "ddd, mmm d, yyyy": "date",
        "dddd, mmmm d, yyyy": "date",
        "d, mmmm yyyy": "date",
    }

    date_output = "%d/%m/%Y"

    def __init__(self):
        self.datebase = self.unixepoch
        self.printformats = []
        self.currentformat = self.currenttype = ""
        self.curx = self.cury = 0
        self.data = Table()
        self.unknown = {}

    def escape(self, s):
        """
        Escape a string
        """
        if s[0:1] == '"':
            return '"' + re.sub('"', '\\"\\"', s[1:-1]) + '"'
        return s

    def parse(self, stream):
        """
        Parse the given stream
        """
        lines = re.sub("[\r\n]+", "\n", stream.read()).split("\n")
        for line in lines:
            self.parseline(line)

    def stream_rows(self):
        """
        Stream the rows (to be used to write a csv file for example)
        """
        return self.data

    def addunknown(self, fld, subfld):
        self.unknown[fld] = self.unknown.get(fld, {})
        self.unknown[fld][subfld] = 1

    def writeunknown(self, stream):
        if self.unknown:
            stream.write("Unrecognized fields (subfields):\n")
            for key in list(self.unknown.keys()):
                stream.write("%s (%s)\n" % (key, repr(list(self.unknown[key].keys()))))
        else:
            stream.write("No unrecognized fields\n")
        stream.flush()

    def _id_field(self, fields):
        if fields[1][:6] in ("PClari", "PApple", "P Sage"):
            self.datebase = self.macepoch

    def _f_field(self, fields):
        for f in fields[1:]:
            ftd = f[0]
            val = f[1:]
            if ftd == "X":
                self.curx = int(val)
            elif ftd == "Y":
                self.cury = int(val)
            elif ftd == "P":
                # references print format for the next cell
                self.currentformat, self.currenttype = self.printformats[int(val)]
            else:
                self.addunknown("F", ftd)

    def _c_field(self, fields):
        for f in fields[1:]:
            ftd = f[0]
            val = f[1:]
            if ftd == "X":
                self.curx = int(val)

            elif ftd == "Y":
                self.cury = int(val)

            elif ftd == "K":
                val = literal_eval(self.escape(val))
                # if type(val) == int:
                #     if self.currenttype == "date":
                #         # value is offset in days from datebase
                #         date = time.localtime(
                #             time.mktime(self.datebase) +
                #             float(val) * 24 * 60 * 60
                #         )

                #         val = time.strftime(self.date_output, date)
                self.data[(self.curx, self.cury)] = "%s" % val

            else:
                self.addunknown("C", ftd)

    def _p_fields(self, fields):
        if fields[1][0] == "P":
            # print formats imply data types?
            format = fields[1][1:].replace("\\", "")
            if format in self.knownformats:
                self.printformats.append((format, self.knownformats[format]))
            else:
                # hack to guess type...
                has_y = "y" in format
                has_d = "d" in format
                has_h = "h" in format
                has_z = "0" in format
                has_p = "." in format
                if (has_d or has_y) and has_h:
                    dtype = "datetime"
                elif has_d or has_y:
                    dtype = "date"
                elif has_h:
                    dtype = "time"
                elif has_p and has_z:
                    dtype = "float"
                elif has_z:
                    dtype = "int"
                else:
                    dtype = "string"
                self.printformats.append((format, dtype))
        else:
            self.addunknown("P", fields[1][0])

    def parseline(self, line):
        fields = re.split("(?i);(?=[a-z])", line)
        if fields[0] == "ID":
            self._id_field(fields)
        if fields[0] == "F":
            self._f_field(fields)
        elif fields[0] == "C":
            self._c_field(fields)
        elif fields[0] == "P":
            self._p_fields(fields)
        else:
            fld = fields[0]
            for f in fields[1:]:
                self.addunknown(fld, f)

    def __iter__(self):
        return self.data.__iter__()



from django.contrib import admin
from wbfdm.admin.instruments import InstrumentModelAdmin

from wbportfolio.models import ProductGroup, ProductGroupRepresentant

from .portfolio_relationships import InstrumentPortfolioThroughModelAdmin


@admin.register(ProductGroup)
class ProductGroupAdmin(InstrumentModelAdmin):
    search_fields = ("name", "currency")

    autocomplete_fields = [
        "management_company",
        "depositary",
        "transfer_agent",
        "administrator",
        "investment_manager",
        "auditor",
        "paying_agent",
        "currency",
    ] + InstrumentModelAdmin.autocomplete_fields

    fieldsets = (
        ("Instrument Information", InstrumentModelAdmin.fieldsets[0][1]),
        (
            "Group Information",
            {
                "fields": (
                    ("type", "category", "umbrella"),
                    ("management_company", "depositary", "transfer_agent", "administrator"),
                    ("investment_manager", "auditor", "paying_agent"),
                    ("net_asset_value_computation_method_path", "order_routing_custodian_adapter", "risk_scale"),
                )
            },
        ),
    )

    inlines = [InstrumentPortfolioThroughModelAdmin]


@admin.register(ProductGroupRepresentant)
class ProductGroupRepresentantModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ["representant"]



from django.shortcuts import get_object_or_404
from rest_framework import filters
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.permission.permissions import InternalUserPermissionMixin
from wbcore.utils.views import MergeMixin

from wbportfolio.filters import CustodianFilterSet
from wbportfolio.models import Custodian
from wbportfolio.serializers import (
    CustodianModelSerializer,
    CustodianRepresentationSerializer,
)

from .configs import (
    CustodianButtonConfig,
    CustodianDisplayConfig,
    CustodianEndpointConfig,
    CustodianTitleConfig,
)


class CustodianRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:custodian"

    filter_backends = (
        filters.SearchFilter,
        filters.OrderingFilter,
    )

    queryset = Custodian.objects.all()
    serializer_class = CustodianRepresentationSerializer

    search_fields = ("name",)
    ordering_fields = search_fields
    ordering = ["name"]


class CustodianModelViewSet(MergeMixin, InternalUserPermissionMixin, viewsets.ModelViewSet):
    queryset = Custodian.objects.all()
    filterset_class = CustodianFilterSet
    serializer_class = CustodianModelSerializer

    search_fields = ("name", "mapping")
    ordering_fields = search_fields
    ordering = ["name"]

    display_config_class = CustodianDisplayConfig
    button_config_class = CustodianButtonConfig
    title_config_class = CustodianTitleConfig
    endpoint_config_class = CustodianEndpointConfig

    @action(detail=True, methods=["PATCH"])
    def split(self, request, pk=None):
        if mapping_choice := request.POST.get("mapping_choice", None):
            custodian = get_object_or_404(Custodian, pk=pk)
            custodian.split_off(mapping_choice)
        return Response({"send": True})



# Generated by Django 4.1.9 on 2023-06-23 07:48

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0001_initial"),
        ("wbportfolio", "0002_product_default_sub_account_squashed_0039_alter_assetallocation_company_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="financial_instrument",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="pms_instruments",
                to="wbfdm.instrument",
                verbose_name="Financial Instrument",
            ),
        ),
    ]



from wbcore import serializers as wb_serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.enums import Button
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display import create_simple_display

from wbportfolio.order_routing import ExecutionInstruction


class ExecutionInstructionSerializer(wb_serializers.Serializer):
    execution_instruction = wb_serializers.ChoiceField(
        choices=ExecutionInstruction.choices,
        default=ExecutionInstruction.MARKET_ON_CLOSE.value,
        clear_dependent_fields=False,
    )
    apply_execution_instruction_to_all_orders = wb_serializers.BooleanField(
        default=False, label="Apply Execution instruction to all orders"
    )

    percentage = wb_serializers.FloatField(
        label="Percentage [1% - 25%]",
        percent=True,
        required=False,
        allow_null=True,
        on_unsatisfied_deps="hide",
        depends_on=[
            {
                "field": "execution_instruction",
                "options": {"activates_on": [ExecutionInstruction.IN_LINE_WITH_VOLUME.value]},
            }
        ],
    )
    price = wb_serializers.FloatField(
        label="Price (optional)",
        required=False,
        allow_null=True,
        on_unsatisfied_deps="hide",
        depends_on=[
            {
                "field": "execution_instruction",
                "options": {"activates_on": [ExecutionInstruction.LIMIT_ORDER.value]},
            }
        ],
    )
    good_for_date = wb_serializers.DateField(
        label="Good For Date (optional)",
        required=False,
        allow_null=True,
        on_unsatisfied_deps="hide",
        depends_on=[
            {
                "field": "execution_instruction",
                "options": {"activates_on": [ExecutionInstruction.LIMIT_ORDER.value]},
            }
        ],
    )
    period = wb_serializers.IntegerField(
        label="Period in minutes",
        required=False,
        allow_null=True,
        on_unsatisfied_deps="hide",
        depends_on=[
            {
                "field": "execution_instruction",
                "options": {"activates_on": [ExecutionInstruction.VWAP.value, ExecutionInstruction.TWAP.value]},
            }
        ],
    )
    time = wb_serializers.TimeField(
        label="Time (UTC)",
        format="%H:%M",
        required=False,
        allow_null=True,
        on_unsatisfied_deps="hide",
        depends_on=[
            {
                "field": "execution_instruction",
                "options": {"activates_on": [ExecutionInstruction.VWAP.value, ExecutionInstruction.TWAP.value]},
            }
        ],
    )


class OrderOrderProposalButtonConfig(ButtonViewConfig):
    def get_create_buttons(self):
        return {
            Button.SAVE_AND_CLOSE.value,
        }

    def get_custom_list_instance_buttons(self) -> set:
        return {
            bt.ActionButton(
                method=RequestType.PUT,
                identifiers=("wbportfolio:order",),
                icon=WBIcon.DEAL_MONEY.icon,
                key="execution_instruction",
                label="Change Execution",
                action_label="Execution changed",
                description_fields="<p>Change Execution</p>",
                serializer=ExecutionInstructionSerializer,
                instance_display=create_simple_display(
                    [
                        ["execution_instruction", "execution_instruction"],
                        ["percentage", "percentage"],
                        ["price", "good_for_date"],
                        ["period", "time"],
                        ["apply_execution_instruction_to_all_orders", "."],
                    ]
                ),
            )
        }



from .rebalancing import (
    RebalancingModelRepresentationSerializer,
    RebalancerRepresentationSerializer,
    RebalancerModelSerializer,
)
from .assets import (
    AssetPositionInstrumentModelSerializer,
    AssetPositionModelSerializer,
    AssetPositionPortfolioModelSerializer,
    AssetPositionAggregatedPortfolioModelSerializer,
    CashPositionPortfolioModelSerializer,
)
from .custodians import CustodianModelSerializer, CustodianRepresentationSerializer
from .portfolio_relationship import (
    InstrumentPreferedClassificationThroughProductModelSerializer,
    InstrumentPortfolioThroughModelSerializer,
)
from .portfolios import (
    PortfolioModelSerializer,
    PortfolioPortfolioThroughModelSerializer,
    PortfolioRepresentationSerializer,
)
from .portfolio_swing_pricing import PortfolioSwingPricingModelSerializer
from .portfolio_cash_targets import PortfolioCashTargetModelSerializer
from .positions import AggregatedAssetPositionModelSerializer
from .registers import RegisterModelSerializer, RegisterRepresentationSerializer
from .roles import PortfolioRoleModelSerializer, PortfolioRoleProjectModelSerializer
from .signals import *
from .transactions import *
from .orders import *
from .products import (
    ProductRepresentationSerializer,
    ProductCustomerRepresentationSerializer,
    ProductUnlinkedRepresentationSerializer,
    ProductListModelSerializer,
    ProductModelSerializer,
    ProductCustomerModelSerializer,
    ProductFeesModelSerializer,
)
from .adjustments import AdjustmentModelSerializer
from .product_group import ProductGroupModelSerializer, ProductGroupRepresentationSerializer
from .portfolio_cash_flow import DailyPortfolioCashFlowModelSerializer
from .reconciliations import AccountReconciliationModelSerializer, AccountReconciliationLineModelSerializer



import datetime
import re
from contextlib import suppress

import numpy as np
import pandas as pd

from wbportfolio.models import ProductGroup


def file_name_parse(file_name):
    dates = re.findall(r"([0-9]{4}-[0-9]{2}-[0-9]{2})", file_name)
    isin = re.findall(r"\.([a-zA-Z0-9]*)_", file_name)
    if len(dates) != 2:
        raise ValueError("Not 2 dates found in the filename")
    if len(isin) != 1:
        raise ValueError("Not exactly 1 isin found in the filename")

    return {
        "isin": isin[0],
        "valuation_date": datetime.datetime.strptime(dates[0], "%Y-%m-%d").date(),
        "generation_date": datetime.datetime.strptime(dates[1], "%Y-%m-%d").date(),
    }


def get_exchange(row):
    if str(row["Asset type code"]) == "VMOB" and row["Bloom Ticker"]:
        with suppress(ValueError):
            [_, exchange] = str(row["Bloom Ticker"]).replace(" Equity", "").replace(" EQUITY", "").split(" ")
            return {"bbg_exchange_codes": exchange}
    return None


def get_underlying_quote(row):
    if str(row["Asset type code"]) == "VMOB":
        data = {
            "instrument_type": "equity",
            "currency__key": row["currency__key"],
            "isin": row["Asset Code"] or row["Security code"],
            "name": row["Asset description"],
        }
        if bloomberg_ticker := row.get("Bloom Ticker", None):
            tickers = str(bloomberg_ticker).replace(" Equity", "").replace(" EQUITY", "").split(" ")
            if len(tickers) > 0:
                data["ticker"] = tickers[0]
            if len(tickers) > 1:
                data["exchange"] = {"bbg_exchange_codes": tickers[1]}
        return data
    return None


def parse(import_source):
    # Load file into a CSV DictReader and convert the encoding to latin1 due to hyphonation
    csv_file = import_source.file.open()
    # csv_file = codecs.iterdecode(csv_file, "latin1")

    cols = [0, 2, 4, 6, 7, 8, 10, 13, 27, 28, 64, 67]

    df = pd.read_csv(csv_file, encoding="latin1", usecols=cols)
    df = df.replace([np.inf, -np.inf, np.nan], None)
    columns_map = {
        "Code": "portfolio",
        "underlying_quote": "underlying_quote",
        "exchange__bbg_exchange_codes": "exchange__bbg_exchange_codes",
        "NAV Date": "date",
        "Quantity / Amount": "initial_shares",
        "Price": "initial_price",
        "Ccy": "currency__key",
        "FX Rate": "initial_currency_fx_rate",
        "asset_valuation_date": "asset_valuation_date",
    }
    df = df.rename(columns=columns_map)
    # Substitute the Code for the ProductGroup ID
    product_group_mapping = {
        product_group.identifier: product_group.id
        for product_group in ProductGroup.objects.filter(identifier__in=df["portfolio"].unique())
    }

    df["date"] = df["date"].apply(lambda x: x.replace("/", "-"))
    df["initial_currency_fx_rate"] = df["initial_currency_fx_rate"].apply(lambda x: 1 / x if x else 1).round(14)

    cash_mask = df["Accounting category"].isin(["T111"])
    cash = (
        df.loc[
            cash_mask,
            ["currency__key", "initial_currency_fx_rate", "portfolio", "date", "initial_shares"],
        ]
        .groupby(
            [
                "currency__key",
                "portfolio",
                "date",
            ]
        )
        .agg(
            {
                "initial_currency_fx_rate": "mean",
                "initial_shares": "sum",
            }
        )
        .reset_index()
    ).copy()
    cash["underlying_quote"] = cash["currency__key"].apply(lambda x: {"currency__key": x, "instrument_type": "cash"})
    cash["initial_price"] = 1.0
    # cash_equivalents_mask, all asset type code that match TRES and which don't have accounting category T111
    cash_equivalents_mask = df["Asset type code"].str.match("TRES") & ~df["Accounting category"].str.match("T111")

    cash_equivalents = (
        df.loc[
            cash_equivalents_mask,
            ["currency__key", "initial_currency_fx_rate", "portfolio", "date", "initial_shares"],
        ]
        .groupby(
            [
                "currency__key",
                "portfolio",
                "date",
            ]
        )
        .agg(
            {
                "initial_currency_fx_rate": "mean",
                "initial_shares": "sum",
            }
        )
        .reset_index()
    ).copy()
    cash_equivalents["underlying_quote"] = cash_equivalents["currency__key"].apply(
        lambda x: {"currency__key": x, "instrument_type": "cash_equivalent"}
    )
    cash_equivalents["initial_price"] = 1.0

    # equities = df.loc[df["Accounting category"].str.match("010"), :].copy() # Historically, we filter out equity base on the "accounting category" that matches "010", we had issue with equity having the code "020". We decided to use the column "Asset type code" and filter out with the code "VCOM"
    equities = df.loc[df["Asset type code"].str.match("VMOB"), :].copy()

    if not equities.empty:
        equities["underlying_quote"] = equities.apply(lambda x: get_underlying_quote(x), axis=1)
        equities["exchange"] = equities.apply(lambda x: get_exchange(x), axis=1)
        del equities["Accounting category"]
        del equities["Asset type code"]
        del equities["Bloom Ticker"]
        del equities["Asset description"]
    df = pd.concat([cash, equities, cash_equivalents])
    df["asset_valuation_date"] = df["date"]
    # Rename the columns

    df = df.drop(df.columns.difference(columns_map.values()), axis=1)

    df["is_estimated"] = False
    df = df.replace([np.inf, -np.inf, np.nan], None)
    df = df[df["underlying_quote"].notnull()]

    df["weighting"] = df.initial_currency_fx_rate * df.initial_price * df.initial_shares
    portfolio_weights = df.groupby("portfolio")["weighting"].sum()
    df["weighting"] = df.apply(lambda x: x["weighting"] / portfolio_weights.loc[x["portfolio"]], axis=1)
    df["portfolio"] = df["portfolio"].apply(
        lambda x: {"id": product_group_mapping[str(x)], "instrument_type": "product_group"}
    )
    csv_file.close()
    return {"data": df.to_dict("records")}



from io import BytesIO

import numpy as np
import pandas as pd

from wbportfolio.import_export.utils import get_file_extension


def parse(import_source):
    if get_file_extension(import_source.file) == ".xlsx":
        df = pd.read_excel(BytesIO(import_source.file.read()), engine="openpyxl", sheet_name="Position").dropna(
            how="all"
        )
    elif get_file_extension(import_source.file) == ".xls":
        df = pd.read_excel(BytesIO(import_source.file.read()), sheet_name="Position").dropna(how="all")
    else:
        raise Exception("File not supported")

    xx, yy = np.where(df == "ISIN Code")
    df = df.iloc[xx[0] :, yy[0] :]
    df = df.rename(columns=df.iloc[0]).drop(df.index[0]).dropna(how="all")

    df["Valuation Date"] = pd.to_datetime(df["Valuation Date"])

    data = list()
    for valuation in df.to_dict("records"):
        isin = valuation["ISIN Code"]
        data.append(
            {
                "instrument": {"instrument_type": "product", "isin": isin},
                "date": valuation["Valuation Date"].strftime("%Y-%m-%d"),
                "net_value": round(valuation["Bid"], 6),
                "calculated": False,
            }
        )

    return {"data": data}



from datetime import date

from wbcore import serializers as wb_serializers
from wbcore.serializers import DefaultFromGET

from wbportfolio.models import Portfolio, Rebalancer, RebalancingModel


class RebalancingModelRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = RebalancingModel
        fields = (
            "id",
            "name",
        )


class RebalancerRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:rebalancer-detail")

    class Meta:
        model = Rebalancer
        fields = ("id", "computed_str", "_detail")


class RebalancerModelSerializer(wb_serializers.ModelSerializer):
    _rebalancing_model = RebalancingModelRepresentationSerializer(source="rebalancing_model")
    frequency_repr = wb_serializers.SerializerMethodField()
    portfolio = wb_serializers.PrimaryKeyRelatedField(
        queryset=Portfolio.objects.all(), default=DefaultFromGET("portfolio")
    )
    activation_date = wb_serializers.DateField(default=lambda: date.today())

    rebalancing_dates = wb_serializers.SerializerMethodField(
        label="Next 10 Rebalancing Dates", field_class=wb_serializers.ListField
    )

    def get_rebalancing_dates(self, obj):
        return list(map(lambda d: d.strftime("%Y-%m-%d"), obj.get_rrule(count=10)))

    def get_frequency_repr(self, obj):
        return obj.frequency_repr

    class Meta:
        model = Rebalancer
        fields = (
            "id",
            "portfolio",
            "computed_str",
            "_rebalancing_model",
            "rebalancing_model",
            "apply_order_proposal_automatically",
            "activation_date",
            "frequency",
            "frequency_repr",
            "rebalancing_dates",
        )
